from os import environ
from pathlib import Path
from typing import List

import click
import requests
from boltons.iterutils import chunked
from rich.panel import Panel
from rich.progress import track

from _incydr_cli import console
from _incydr_cli import logging_options
from _incydr_cli import render
from _incydr_cli.cmds.models import AgentCSV
from _incydr_cli.cmds.models import AgentJSON
from _incydr_cli.cmds.options.output_options import columns_option
from _incydr_cli.cmds.options.output_options import input_format_option
from _incydr_cli.cmds.options.output_options import single_format_option
from _incydr_cli.cmds.options.output_options import SingleFormat
from _incydr_cli.cmds.options.output_options import table_format_option
from _incydr_cli.cmds.options.output_options import TableFormat
from _incydr_cli.core import incompatible_with
from _incydr_cli.core import IncydrCommand
from _incydr_cli.core import IncydrGroup
from _incydr_sdk.agents.models import Agent
from _incydr_sdk.core.client import Client
from _incydr_sdk.utils import model_as_card


@click.group(cls=IncydrGroup)
@logging_options
def agents():
    """
    View and manage Incydr agents.

    Incydr agents run on the endpoints in your environment and monitor for insider risk activity.
    """


@agents.command("list", cls=IncydrCommand)
@click.option(
    "--active/--inactive",
    default=None,
    help="Filter by active or inactive agents. Defaults to returning both when when neither option is passed.",
)
@click.option(
    "--healthy",
    is_flag=True,
    default=None,
    help="Filter by healthy agents. Agents that have no health issue types are considered healthy.",
    cls=incompatible_with("unhealthy"),
)
@click.option(
    "--unhealthy",
    is_flag=False,  # is_flag=False with a flag_value indicates an optional value
    flag_value="FLAG_VALUE",
    default=None,
    help="Filter by unhealthy agents. Defaults to returning all unhealthy agents."
    " Pass a comma delimited list of health issue types to filter by unhealthy agents that have (at least) any "
    "of the given health issue type(s). Health issue types include the following: NOT_CONNECTING, NOT_SENDING_SECURITY_EVENTS, SECURITY_INGEST_REJECTED, MISSING_MACOS_PERMISSION_FULL_DISK_ACCESS, MISSING_MACOS_PERMISSION_ACCESSIBILITY.",
    cls=incompatible_with("healthy"),
)
@table_format_option
@columns_option
@logging_options
def list_(
    active: bool = None,
    healthy: bool = None,
    unhealthy: str = None,
    format_: TableFormat = None,
    columns: str = None,
):
    """
    List agents.
    """
    agent_healthy = None
    health_issues = None
    if healthy:
        agent_healthy = True
    if unhealthy:
        agent_healthy = False
        if (
            not unhealthy == "FLAG_VALUE"
        ):  # If the unhealthy value is FLAG_VALUE then we know the option was passed with no values
            health_issues = unhealthy.split(",")

    client = Client()

    agents = client.agents.v1.iter_all(
        active=active,
        agent_healthy=agent_healthy,
        agent_health_issue_types=health_issues,
    )

    if format_ == TableFormat.table:
        render.table(Agent, agents, columns=columns, flat=False)
    elif format_ == TableFormat.csv:
        render.csv(Agent, agents, columns=columns, flat=True)
    elif format_ == TableFormat.json_pretty:
        for agent in agents:
            console.print_json(agent.json())
    else:
        for agent in agents:
            click.echo(agent.json())


@agents.command(cls=IncydrCommand)
@click.argument("agent_id")
@single_format_option
@logging_options
def show(
    agent_id: str,
    format_: SingleFormat,
):
    """
    Show details for a single agent.
    """
    client = Client()
    agent = client.agents.v1.get_agent(agent_id)

    if format_ == SingleFormat.rich and client.settings.use_rich:
        console.print(Panel.fit(model_as_card(agent), title=f"Agent {agent.agent_id}"))
    elif format_ == SingleFormat.json_pretty:
        console.print_json(agent.json())
    else:
        click.echo(agent.json())


@agents.command(cls=IncydrCommand)
@click.argument("file", type=click.File())
@input_format_option
@logging_options
def bulk_activate(file: Path, format_: str):
    """
    Activate a group of agents from a file (CSV or JSON-LINES formatted).

    \b
    Use `-` as filename to read from stdin.

    Input files require a header (for CSV input) or JSON key for each object (for JSON-LINES input) to identify
    which agent ID to activate.

    Header and JSON key values that are accepted are: agentGuid, agent_id, agentId, or guid
    """
    chunk_size = (
        environ.get("incydr_batch_size") or environ.get("INCYDR_BATCH_SIZE") or 50
    )
    try:
        chunk_size = int(chunk_size)
    except ValueError:
        console.print(
            f"INCYDR_BATCH_SIZE environment variable must be an integer, found: '{chunk_size}'"
        )
        return

    # parse CSV or JSON input
    if format_ == "csv":
        models = AgentCSV.parse_csv(file)
    else:
        models = AgentJSON.parse_json_lines(file)
    try:
        agent_ids = [agent.agent_id for agent in models]
    except ValueError as err:
        console.print(err)
        return

    # validate we got at least one agent_id
    num_agents = len(agent_ids)
    if num_agents < 1:
        console.print(f"[red]No agent IDs found in {format_} input.")
        return

    client = Client()
    batches = chunked(agent_ids, size=chunk_size)
    for batch in track(batches, description="Activating agents...", console=console):
        process_batch(client, batch, activate=True)


@agents.command(cls=IncydrCommand)
@click.argument("file", type=click.File())
@input_format_option
@logging_options
def bulk_deactivate(file: Path, format_: str):
    """
    Deactivate a group of agents from a file (CSV or JSON-LINES formatted).

    \b
    Use `-` as filename to read from stdin.

    Input files require a header (for CSV input) or JSON key for each object (for JSON-LINES input) to identify
    which agent ID to deactivate.

    Header and JSON key values that are accepted are: agentGuid, agent_id, agentId, or guid
    """
    chunk_size = (
        environ.get("incydr_batch_size") or environ.get("INCYDR_BATCH_SIZE") or 50
    )
    try:
        chunk_size = int(chunk_size)
    except ValueError:
        console.print(
            f"INCYDR_BATCH_SIZE environment variable must be an integer, found: '{chunk_size}'"
        )
        return

    # parse CSV or JSON input
    if format_ == "csv":
        models = AgentCSV.parse_csv(file)
    else:
        models = AgentJSON.parse_json_lines(file)
    try:
        agent_ids = [agent.agent_id for agent in models]
    except ValueError as err:
        console.print(err)
        return

    # validate we got at least one agent_id
    num_agents = len(agent_ids)
    if num_agents < 1:
        console.print(f"[red]No agent IDs found in {format_} input.")
        return

    client = Client()
    batches = chunked(agent_ids, size=chunk_size)
    for batch in track(batches, description="Deactivating agents...", console=console):
        process_batch(client, batch, activate=False)


def process_batch(client: Client, batch: List[str], activate: bool):
    action = "activation" if activate else "deactivation"
    api_call = client.agents.v1.activate if activate else client.agents.v1.deactivate
    process_individually = False
    try:
        api_call(batch)
    except requests.HTTPError as err:
        if err.response.status_code == 404:
            invalid_agent_ids = err.response.json().get("agentsNotFound")
            if invalid_agent_ids is None:
                console.print(
                    f"[red]Unknown 404 error processing batch of {len(batch)} agent {action}s."
                )
                process_individually = True
            else:
                console.print(
                    f"[red]404 Error processing batch of {len(batch)} agent {action}s, agent_ids not found:[/red] {invalid_agent_ids}"
                )
                batch = list(set(batch) - set(invalid_agent_ids))
                if len(batch) > 0:
                    console.print("Removing invalid agent_ids and retrying...")
                    try:
                        api_call(batch)
                    except requests.HTTPError as err:
                        console.print(f"[red]Error retrying batch. {err.response.text}")
                        process_individually = True
        else:
            console.print(
                f"[red]Unknown error processing batch of {len(batch)} agent {action}s."
            )
            process_individually = True
    if process_individually and len(batch) > 1:
        console.print(f"Trying agent {action} for this batch individually.")
        for agent_id in batch:
            try:
                api_call(agent_id)
            except requests.HTTPError as err:
                msg = f"Failed to process {action} for {agent_id}: {err.response.text}"
                client.settings.logger.error(msg)
                console.print(msg)



from pydantic import ValidationError


class IncydrException(Exception):
    """Base class for all Incydr specific exceptions."""

    ...


class AuthMissingError(ValidationError, IncydrException):
    def __init__(self, validation_error: ValidationError):
        self.pydantic_error = str(validation_error)
        super().__init__(validation_error.raw_errors, validation_error.model)

    @property
    def error_keys(self):
        return [e["loc"][0] for e in self.errors()]

    def __str__(self):
        return (
            f"{self.pydantic_error}\n\n"
            "Pass required args to the `incydr.Client` or set required values in your environment.\n\n"
            "See https://developer.code42.com/sdk/settings for more details."
        )


class WatchlistNotFoundError(IncydrException):
    """Raised when a watchlist with the matching type or title is not found."""

    def __init__(self, name):
        self.name = name
        self.message = f"Watchlist Not Found Error: No watchlist matching the type or title '{name}' was found."
        super().__init__(self.message)


class DateParseError(IncydrException):
    """An error raised when the date data cannot be parsed."""

    def __init__(self, date, msg=None):
        self._date = date
        message = msg or f"DateParseError: Error parsing time data: '{date}'."
        super().__init__(message)

    @property
    def date(self):
        return self._date



from __future__ import annotations

from datetime import datetime
from typing import List
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from _incydr_sdk.core.models import ResponseModel
from _incydr_sdk.enums import _Enum


class SortKeys(_Enum):
    """Possible keys to sort agents list results by."""

    NAME = "NAME"
    USER_ID = "USER_ID"
    AGENT_TYPE = "AGENT_TYPE"
    OS_HOSTNAME = "OS_HOSTNAME"
    LAST_CONNECTED = "LAST_CONNECTED"
    OS_NAME = "OS_NAME"


class AgentType(_Enum):
    """Possible types of agents."""

    CODE42AAT = "CODE42AAT"
    COMBINED = "COMBINED"
    CODE42 = "CODE42"


class Agent(ResponseModel):
    """
    A model representing an Incydr agent.

    **Fields**:

    * **agent_id**: `str` The globally unique ID (guid) for this agent.
    * **name**: `str` The editable name of the agent.
    * **user_id**: `str` The unique ID of the user the agent is assigned to.
    * **os_hostname**: `str` The hostname reported by the OS the agent is running on.
    * **os_name**: `str` The name of the OS the agent is running on.
    * **machine_id**: `str` Device machine ID.
    * **serial_number**: `str` Authenticated agent serial number.
    * **active**: `bool` If the agent status is active.
    * **agent_type**: [`AgentType`][agent-type] The type of agent.
    * **agent_health_issue_types: `List[str]` List of health issues with the agent. Health issue types include the following: `NOT_CONNECTING`, `NOT_SENDING_SECURITY_EVENTS`, `SECURITY_INGEST_REJECTED`, `MISSING_MACOS_PERMISSION_FULL_DISK_ACCESS`, `MISSING_MACOS_PERMISSION_ACCESSIBILITY`.
    * **app_version**: `str` The app version of the agent.
    * **product_version**: `str` The product version of the agent.
    * **last_connected**: `datetime` The time the agent last connected to a Code42 Authority server.
    * **external_reference**: `str` Editable reference information (useful for identifying an agent in external systems).
    * **creation_date**: `datetime` The time the agent was first registered.
    * **modification_date**: `datetime` The time the agent's database entry was last updated.
    """

    agent_id: Optional[str] = Field(alias="agentId")
    name: Optional[str]
    user_id: Optional[str] = Field(alias="userId")
    os_hostname: Optional[str] = Field(alias="osHostname")
    os_name: Optional[str] = Field(alias="osName")
    machine_id: Optional[str] = Field(alias="machineId")
    serial_number: Optional[str] = Field(alias="serialNumber")
    active: Optional[bool]
    agent_type: Optional[Union[AgentType, str]] = Field(alias="agentType")
    agent_health_issue_types: Optional[List[str]] = Field(alias="agentHealthIssueTypes")
    app_version: Optional[str] = Field(alias="appVersion")
    product_version: Optional[str] = Field(alias="productVersion")
    last_connected: Optional[datetime] = Field(alias="lastConnected")
    external_reference: Optional[str] = Field(alias="externalReference")
    creation_date: Optional[datetime] = Field(alias="creationDate")
    modification_date: Optional[datetime] = Field(alias="modificationDate")


class AgentsPage(ResponseModel):
    """
    A model representing a page of Agents.

    **Fields**:

    * **agents**: `List[[Agent][agent-model]]` The list of agents returned from the query.
    * **total_count**: `int` Total number of agents found in query results.
    * **page_size**: `int` The maximum number of agents returned in query results page.
    * **page_num**: `int` The current page number of the query result set.
    """

    agents: List[Agent]
    total_count: int = Field(alias="totalCount")
    page_size: int = Field(alias="pageSize")
    page_num: int = Field(alias="page")


class AgentUpdateRequest(BaseModel):
    name: str
    externalReference: str


class QueryAgentsRequest(BaseModel):
    active: Optional[bool]
    agentType: Optional[Union[AgentType, str]]
    agentHealthy: Optional[bool]
    anyOfAgentHealthIssueTypes: Optional[List[str]]
    srtKey: Optional[Union[SortKeys, str]]
    srtDir: Optional[str]
    pageSize: Optional[int]
    page: Optional[int]

    @validator("srtDir")
    def _validate(cls, value):  # noqa
        value = str(value).upper()
        assert value in ("ASC", "DESC")
        return value



from itertools import count

from _incydr_sdk.departments.models import DepartmentsPage
from _incydr_sdk.departments.models import GetPageRequest


class DepartmentsClient:
    def __init__(self, parent):
        self._parent = parent
        self._v1 = None

    @property
    def v1(self):
        if self._v1 is None:
            self._v1 = DepartmentsV1(self._parent)
        return self._v1


class DepartmentsV1:
    """
    Client for `/v1/departments` endpoints.

    Usage example:

        >>> import incydr
        >>> client = incydr.Client(**kwargs)
        >>> client.departments.v1.get_page()
    """

    def __init__(self, parent):
        self._parent = parent

    def get_page(self, page_num: int = 1, page_size=None, name=None) -> DepartmentsPage:
        """
        Get a page of departments.  Retrieves department information that has been pushed to Code42 from SCIM or User Directory Sync.
        The resulting department names can be used to include departments on watchlists.

        **Parameters**:

        * **page_num**: `int` - Page number for results, starting at 1.
        * **page_size**: `int` - Max number of results to return for a page.
        * **user_id**: `str` - Matches departments whose name is like the given value.

        **Returns**: A [`DepartmentsPage`][departmentspage-model] object.
        """
        data = GetPageRequest(
            page=page_num,
            page_size=page_size or self._parent.settings.page_size,
            name=name,
        )
        response = self._parent.session.get("/v1/departments", params=data.dict())
        return DepartmentsPage.parse_response(response)

    def iter_all(self, page_size=None, name=None):
        """
        Iterate over all departments.

        Accepts the same parameters as `.get_page()` excepting `page_num`.

        **Returns**: A generator yielding individual department names (`str`).
        """
        page_size = page_size or self._parent.settings.page_size
        for page_num in count(1):
            page = self.get_page(page_num=page_num, page_size=page_size, name=name)
            yield from page.departments
            if len(page.departments) < page_size:
                break



from datetime import datetime
from itertools import count
from pathlib import Path
from typing import Iterator
from typing import List
from typing import Tuple
from typing import Union

from requests import Response

from _incydr_sdk.cases.models import Case
from _incydr_sdk.cases.models import CaseDetail
from _incydr_sdk.cases.models import CaseFileEvents
from _incydr_sdk.cases.models import CasesPage
from _incydr_sdk.cases.models import CreateCaseRequest
from _incydr_sdk.cases.models import QueryCasesRequest
from _incydr_sdk.cases.models import UpdateCaseRequest
from _incydr_sdk.core.utils import get_filename_from_content_disposition
from _incydr_sdk.enums import SortDirection
from _incydr_sdk.enums.cases import CaseStatus
from _incydr_sdk.enums.cases import SortKeys
from _incydr_sdk.file_events.models.event import FileEventV2
from _incydr_sdk.queries.utils import MICROSECOND_FORMAT


class CasesV1:
    """
    Client for `/v1/cases` endpoints.

    Usage example:

        >>> import incydr
        >>> client = incydr.Client(**kwargs)
        >>> client.cases.v1.get_case(23)
    """

    def __init__(self, parent):
        self._parent = parent

    def create(
        self,
        name: str,
        subject: str = None,
        assignee: str = None,
        description: str = None,
        findings: str = None,
    ) -> CaseDetail:
        """
        Create a case.

        **Parameters:**

        * **name**: `str` (required) The unique name given to the case.
        * **subject**: `str` The user UID of the subject being investigated in this case.
        * **assignee**: `str` The actor ID of the administrator assigned to investigate the case.
        * **findings**: `str` Markdown formatted text summarizing the findings for a case.
        * **description**: `str` Brief description providing context for a case.

        **Returns**: A [`Case`][case-model] object representing the newly created case.
        """
        data = CreateCaseRequest(
            name=name,
            subject=subject,
            assignee=assignee,
            findings=findings,
            description=description,
        )
        response = self._parent.session.post(url="/v1/cases", json=data.dict())
        return CaseDetail.parse_response(response)

    def delete(self, case_number: Union[int, Case]) -> Response:
        """
        Delete a case.

        **Parameters**:

        * **case_number** `int | Case` Unique numeric identifier for the case or a [`Case`](../models/#case) object.

        Usage example:

            >>> client.cases.v1.delete(23)
            <Response [204]>

            # Alternatively:
            >>> case = client.cases.v1.get_case(23)
            >>> client.cases.v1.delete(case)

        **Returns**: A `requests.Response` indicating success.
        """
        if isinstance(case_number, Case):
            case_number = case_number.number
        return self._parent.session.delete(f"/v1/cases/{case_number}")

    def get_case(self, case_number: int) -> CaseDetail:
        """
        Get a single case.

        **Parameters**:

        * **case_number**: `int` Unique numeric identifier for the case.

        **Returns**: A [`Case`][case-model] object representing the case.
        """
        response = self._parent.session.get(f"/v1/cases/{case_number}")
        return CaseDetail.parse_response(response)

    def get_page(
        self,
        assignee: str = None,
        created_at: Tuple[datetime, datetime] = None,
        is_assigned: bool = None,
        last_modified_by: str = None,
        name: str = None,
        status: CaseStatus = None,
        page_num: int = 1,
        page_size: int = None,
        sort_dir: SortDirection = SortDirection.ASC,
        sort_key: SortKeys = SortKeys.NUMBER,
    ) -> CasesPage:
        """
        Get a page of cases.

        Filter results by passing appropriate parameters:

        **Parameters**:

        * **assignee**: `str` - Actor ID of an assignee of a case on which to filter.
        * **created_at**: `Tuple[datetime, datetime]` - Filter cases created between the supplied start and end times.
        * **is_assigned**: `bool` - Filter cases with an assignee (`True`) or without (`False`).
        * **last_modified_by**: `str` - User UID of the user who most recently modified the case.
        * **name**: str - Name of a case on which to filter; will include partial matches.
        * **status**: [`CaseStatus`][case-statuses] - One or more case statuses on which to filter. Available values: `OPEN`, `CLOSED`
        * **page_num**: `int` - Page number for results, starting at 1.
        * **page_size**: `int` - Max number of results to return for a page. Defaults to client's `page_size` setting.
        * **sort_dir**: `SortDirection` - The direction on which to sort the response, based on the corresponding key.
        * **sort_key**: [`SortKeys`][cases-sort-keys] - One or more values on which the response will be sorted.

        **Returns**: A [`CasesPage`][casespage-model] object.
        """

        if created_at is not None:
            if not isinstance(created_at, (Tuple, list)):
                raise TypeError(
                    f"created_at kwarg should be a Tuple[datetime, datetime] object"
                    f", passed 'created_at={created_at}' of type: {type(created_at)}"
                )
            else:
                created_at = f"{created_at[0].strftime(MICROSECOND_FORMAT)}/{created_at[1].strftime(MICROSECOND_FORMAT)}"

        data = QueryCasesRequest(
            assignee=assignee,
            createdAt=created_at,
            isAssigned=is_assigned,
            lastModifiedBy=last_modified_by,
            name=name,
            status=status,
            pgNum=page_num,
            pgSize=page_size or self._parent.settings.page_size,
            srtDir=sort_dir,
            srtKey=sort_key,
        )
        response = self._parent.session.get("/v1/cases", params=data.dict())
        return CasesPage.parse_response(response)

    def iter_all(
        self,
        assignee: str = None,
        created_at: Tuple[datetime, datetime] = None,
        is_assigned: bool = None,
        last_modified_by: str = None,
        name: str = None,
        status: CaseStatus = None,
        page_size: int = None,
        sort_dir: SortDirection = SortDirection.ASC,
        sort_key: SortKeys = SortKeys.NUMBER,
    ) -> Iterator[Case]:
        """
        Iterate over all cases.

        Accepts the same parameters as `.get_page()` excepting `page_num`.

        **Returns**: A generator yielding individual [`Case`][case-model] objects.
        """
        page_size = page_size or self._parent.settings.page_size
        for page_num in count(1):
            page = self.get_page(
                assignee=assignee,
                created_at=created_at,
                is_assigned=is_assigned,
                last_modified_by=last_modified_by,
                name=name,
                status=status,
                page_num=page_num,
                page_size=page_size,
                sort_dir=sort_dir,
                sort_key=sort_key,
            )
            yield from page.cases
            if len(page.cases) < page_size:
                break

    def update(self, case: Union[Case, CaseDetail]):
        """
        Updates a case.
        Valid updatable fields: name, assignee, description, findings, subject, status

        **Parameters**

        * **case**: [`Case`][case-model] The modified case object.

        Usage example:

            >>> case = client.cases.v1.get_case(23)
            >>> case.name = "Updated name"
            >>> client.cases.v1.update(case)

        **Returns**: A [`Case`][case-model] object with updated values from server.
        """
        data = UpdateCaseRequest(**case.dict())
        response = self._parent.session.put(
            f"/v1/cases/{case.number}", json=data.dict()
        )
        return CaseDetail.parse_response(response)

    def download_summary_pdf(
        self, case_number: int, target_folder: Union[str, Path]
    ) -> Path:
        """
        Downloads a summary of a case in PDF format to specified target folder.

        **Parameters**

        * **case_number**: `int` Unique numeric identifier for the case.
        * **target_folder**: `Path | str` A string or `pathlib.Path` object that represents the folder which the PDF
            will be saved to.

        **Returns**: A `pathlib.Path` object representing location of the downloaded PDF.
        """
        folder = Path(target_folder)  # ensure a Path object if we get passed a string
        if not folder.is_dir():
            raise ValueError(
                f"`target_folder` argument must resolve to a folder: {target_folder}"
            )
        response = self._parent.session.get(f"/v1/cases/{case_number}/export")
        filename = get_filename_from_content_disposition(
            response, fallback=f"Case-{case_number}.pdf"
        )
        target = folder / filename
        target.write_bytes(response.content)
        return target

    def download_file_event_csv(self, case_number: int, target_folder: Path) -> Path:
        """
        Downloads all file event data for a case in CSV format to specified target folder.

        **Parameters**

        * **case_number**: `int` Unique numeric identifier for the case.
        * **target_folder**: `Path | str` A string or `pathlib.Path` object that represents the folder which the CSV
            will be saved to.

        **Returns**: A `pathlib.Path` object representing location of the downloaded CSV.
        """
        folder = Path(target_folder)  # ensure a Path object if we get passed a string
        if not folder.is_dir():
            raise ValueError(
                f"`target_folder` argument must resolve to a folder: {target_folder}"
            )
        response = self._parent.session.get(f"/v1/cases/{case_number}/fileevent/export")
        filename = get_filename_from_content_disposition(
            response, fallback=f"Case-{case_number}-file-events.csv"
        )
        target = folder / filename
        target.write_bytes(response.content)
        return target

    def download_full_case_zip(
        self,
        case_number,
        target_folder: Path,
        include_files=True,
        include_summary=True,
        include_file_events=True,
    ) -> Path:
        """
        Downloads full export of case in ZIP format to specified target folder.

        **Parameters**

        * **case_number**: `int` Unique numeric identifier for the case.
        * **target_folder**: `Path | str` A string or `pathlib.Path` object that represents the folder which the ZIP
            will be saved to.
        * **include_files**: `bool` Include source files (if they are available) in the .zip. default=True
        * **include_summary**: `bool` Include summary PDF in the .zip. default=True
        * **include_file_events**: `bool` Include file events .csv in the .zip. default=True

        **Returns**: A `pathlib.Path` object representing location of the downloaded ZIP.
        """
        folder = Path(target_folder)  # ensure a Path object if we get passed a string
        response = self._parent.session.get(
            f"/v1/cases/{case_number}/export/full",
            params={
                "files": include_files,
                "summary": include_summary,
                "fileEvents": include_file_events,
            },
        )
        filename = get_filename_from_content_disposition(
            response, fallback=f"Case-{case_number}.zip"
        )
        target = folder / filename
        target.write_bytes(response.content)
        return target

    def download_file_for_event(
        self, case_number: int, event_id: str, target_folder: Path
    ) -> Path:
        """
        Download the source file (if captured) from a file event attached to a case.

        **Parameters**

        * **case_number**: `int` Unique numeric identifier for the case.
        * **event_id**: `str` Unique identifier for event to download.
        * **target_folder**: `Path | str` A string or `pathlib.Path` object that represents the folder which the file
            will be saved to.

        **Returns**: A `pathlib.Path` object representing location of the downloaded file.
        """
        folder = Path(target_folder)  # ensure a Path object if we get passed a string
        response = self._parent.session.get(
            f"/v1/cases/{case_number}/fileevent/{event_id}/file"
        )
        filename = get_filename_from_content_disposition(
            response, fallback=f"Case-{case_number}-{event_id}-unknown-filename"
        )
        target = folder / filename
        target.write_bytes(response.content)
        return target

    def get_file_events(
        self, case_number: int, page_num: int = 1, page_size: int = None
    ) -> CaseFileEvents:
        """
        Get abbreviated details for file events attached to a case.

        **Parameters**

        * **case_number**: `int` - Unique numeric identifier for the case.
        * **page_num**: `int` - Page number for results, starting at 1.
        * **page_size**: `int` - Max number of results to return for a page. Defaults to client's `page_size` setting.

        **Returns**: A [`CaseFileEvents`][casefileevents-model] object containing the associated file events.
        """
        params = {
            "pgNum": page_num,
            "pgSize": page_size or self._parent.settings.page_size,
        }
        r = self._parent.session.get(
            f"/v1/cases/{case_number}/fileevent", params=params
        )
        return CaseFileEvents.parse_response(r)

    def add_file_events_to_case(
        self, case_number: int, event_ids: Union[str, List[str]]
    ) -> Response:
        """
        Attach file events to a case.

        **Parameters:**

        * **case_number**: `int` Unique numeric identifier for the case.
        * **event_ids**: `str | List[str]` A string or list of strings representing the eventId(s) to attach to the
            case.

        **Returns**: A `requests.Response` indicating success.
        """
        if isinstance(event_ids, str):
            event_ids = [event_ids]
        return self._parent.session.post(
            f"/v1/cases/{case_number}/fileevent", json={"events": event_ids}
        )

    def delete_file_event_from_case(self, case_number: int, event_id: str) -> Response:
        """
        Remove file events from a case.

        **Parameters:**

        * **case_number**: `int` Unique numeric identifier for the case.
        * **event_id**: `str` Unique identifier of event to remove from the case.

        **Returns**: A `requests.Response` indicating success.
        """
        return self._parent.session.delete(
            f"/v1/cases/{case_number}/fileevent/{event_id}"
        )

    def get_file_event_detail(self, case_number: int, event_id: str):
        """
        Get the full detail for a given file event attached to a case.

        **Parameters**

        * **case_number**: `int` Unique numeric identifier for the case.
        * **event_id**: `str` Unique identifier for event associated with case.

        **Returns**:  A [`FileEventV2`][fileevent-model] object representing the file event.
        """
        response = self._parent.session.get(
            f"/v1/cases/{case_number}/fileevent/{event_id}"
        )
        return FileEventV2.parse_response(response)


class CasesClient:
    def __init__(self, parent):
        self._parent = parent
        self._v1 = None

    @property
    def v1(self):
        if self._v1 is None:
            self._v1 = CasesV1(self._parent)
        return self._v1



from typing import Optional

from pydantic import SecretStr
from requests import Session
from requests.auth import AuthBase
from requests.auth import HTTPBasicAuth

from _incydr_sdk.core.models import AuthResponse


class APIClientAuth(AuthBase):
    def __init__(
        self, session: Session, api_client_id: str, api_client_secret: SecretStr
    ):
        self.session = session
        self.api_client_id = api_client_id
        self.api_client_secret = api_client_secret
        self.token_response: Optional[AuthResponse] = None

    def refresh(self):
        auth = HTTPBasicAuth(
            username=self.api_client_id,
            password=self.api_client_secret.get_secret_value(),
        )
        r = self.session.post("/v1/oauth", auth=auth)
        r.raise_for_status()
        self.token_response = AuthResponse.parse_response(r)

    def __call__(self, request):
        if self.token_response is None or self.token_response.expired:
            self.refresh()
        token = self.token_response.access_token.get_secret_value()
        request.headers["Authorization"] = f"Bearer {token}"
        return request



import logging
from threading import Lock

import click
from click import BadOptionUsage

from _incydr_cli.logger.enums import ServerProtocol
from _incydr_cli.logger.handlers import NoPrioritySysLogHandler

# prevent loggers from printing stacks to stderr if a pipe is broken


logging.raiseExceptions = False

logger_deps_lock = Lock()


def _init_logger(logger, handler):
    logger.setLevel(logging.INFO)
    handler.setFormatter(logging.Formatter(fmt="%(message)s"))
    logger.addHandler(handler)
    return logger


def get_logger_for_server(protocol, hostname, port, certs):
    """Gets the logger that sends logs to a server..

    Args:
        hostname: The hostname of the server. It may include the port.
        port: The port for the server.  Defaults to 601.
        protocol: The transfer protocol for sending logs.
        certs: Use for passing SSL/TLS certificates when connecting to the server.
    """
    logger = logging.getLogger("incydr_syslog")
    if len(logger.handlers):  # if logger has handlers
        return logger

    with logger_deps_lock:
        hostname = hostname
        port = port or 601
        protocol = protocol.value if isinstance(protocol, ServerProtocol) else protocol
        if not len(logger.handlers):
            handler = NoPrioritySysLogHandler(hostname, int(port), protocol, certs)
            handler.connect_socket()
            return _init_logger(logger, handler)
    return logger


def get_server_logger(output, certs, ignore_cert_validation):

    # parse output
    output = output.split(":")

    protocol = ServerProtocol.TCP
    port = 601

    if len(output) == 1:  # HOSTNAME
        hostname = output[0]
    elif len(output) == 2:  # HOSTNAME:PORT
        hostname = output[0]
        port = output[1]
    elif len(output) == 3:  # PROTOCOL:HOSTNAME:PORT
        protocol = output[0].upper()
        hostname = output[1]
        port = output[2]
    else:
        raise BadOptionUsage(
            "output",
            "Error parsing output string.  Pass a string in the format PROTOCOL:HOSTNAME:PORT to output "
            "to the specified server endpoint, where format is either TCP, TLS-TCP, or UDP. "
            "Also accepts strings of the format HOSTNAME and HOSTNAME:PORT where port will default to 601 "
            "and protocol will default to TCP.",
        )

    # certs and ignore-cert-validation only compatible with TLS-TCP
    if protocol != ServerProtocol.TLS_TCP:
        arg = None
        if ignore_cert_validation:
            arg = "ignore-cert-validation"
        elif certs is not None:
            arg = "certs"
        if arg is not None:
            raise click.BadOptionUsage(
                arg,
                f"'{arg}' can only be used with '{ServerProtocol.TLS_TCP}' protocol.",
            )

    if ignore_cert_validation:
        certs = "ignore"

    try:
        return get_logger_for_server(protocol, hostname, port, certs)
    except Exception as err:
        raise ConnectionError(
            f"Unable to connect to {hostname}. Failed with error: {err}."
        )



import json
from contextlib import nullcontext
from pathlib import Path
from typing import List
from typing import Optional

import click
from pydantic import Field
from requests import HTTPError
from rich.panel import Panel
from rich.progress import track

from _incydr_cli import console
from _incydr_cli import get_user_project_path
from _incydr_cli import logging_options
from _incydr_cli import render
from _incydr_cli.cmds.options.output_options import columns_option
from _incydr_cli.cmds.options.output_options import input_format_option
from _incydr_cli.cmds.options.output_options import output_options
from _incydr_cli.cmds.options.output_options import single_format_option
from _incydr_cli.cmds.options.output_options import SingleFormat
from _incydr_cli.cmds.options.output_options import table_format_option
from _incydr_cli.cmds.options.output_options import TableFormat
from _incydr_cli.cmds.options.utils import checkpoint_option
from _incydr_cli.cmds.utils import warn_interrupt
from _incydr_cli.core import IncydrCommand
from _incydr_cli.core import IncydrGroup
from _incydr_cli.cursor import CursorStore
from _incydr_cli.logger import get_server_logger
from _incydr_sdk.core.client import Client
from _incydr_sdk.core.models import CSVModel
from _incydr_sdk.core.models import Model
from _incydr_sdk.enums.sessions import ContentInspectionStatuses
from _incydr_sdk.enums.sessions import SessionSeverities
from _incydr_sdk.enums.sessions import SessionStates
from _incydr_sdk.file_events.models.event import FileEventV2
from _incydr_sdk.file_events.models.response import FileEventsPage
from _incydr_sdk.sessions.models.response import Session
from _incydr_sdk.utils import model_as_card

SEVERITY_MAP = {"NO_RISK": 0, "LOW": 1, "MODERATE": 2, "HIGH": 3, "CRITICAL": 4}


@click.group(cls=IncydrGroup)
@logging_options
def sessions():
    """
    View and manage Incydr Sessions, including Alerts.

    Sessions group all Incydr activity for an individual actor within a specific time frame.
    Sessions may include file events, alerts, Instructor lessons, and more. A single session can contain one or more alerts.

    Alerts are triggered when user file activity matches criteria defined in your alert rules.
    Alerts are included within sessions to group related activity and provide additional context.
    """


@sessions.command(cls=IncydrCommand)
@checkpoint_option
@click.option(
    "--actor-id", default=None, help="Limit search to sessions generated by this actor."
)
@click.option(
    "--start",
    default=None,
    help="Limit search to sessions beginning on or after this date and time. "
    "Accepts a date/time in yyyy-MM-dd (UTC) or yyyy-MM-dd HH:MM:SS (UTC+24-hr time) format.",
)
@click.option(
    "--end",
    default=None,
    help="Limit search to sessions beginning before this date and time. "
    "Accepts a date/time in yyyy-MM-dd (UTC) or yyyy-MM-dd HH:MM:SS (UTC+24-hr time) format.",
)
@click.option(
    "--no-alerts",
    is_flag=True,
    default=False,
    help="Limit search to sessions that do NOT have alerts associated with them.",
)
@click.option(
    "--risk-indicators",
    default=None,
    help="A CSV list of risk indicator IDs. Limit search to sessions that include these risk indicators.",
)
@click.option(
    "--state",
    type=click.Choice([e.value for e in SessionStates]),
    default=[],
    multiple=True,
    help="Limit search to sessions with this state. Can be specified multiple times to include multiple values. ",
)
@click.option(
    "--severity",
    type=click.Choice([e.value for e in SessionSeverities]),
    default=[],
    multiple=True,
    help="Specify one or more severity levels, can be specified multiple times to include multiple values. "
    "Limit search to sessions that have the matching severity value(s). ",
)
@click.option(
    "--rule-id",
    default=[],
    multiple=True,
    help="A rule ID. Limit search to sessions associated with this rule. Can be specified multiple times to include multiple values. ",
)
@click.option(
    "--watchlist-id",
    default=[],
    multiple=True,
    help="A watchlist ID. Limit search to sessions associated with this watchlist. Can be specified multiple times to include multiple values. ",
)
@click.option(
    "--content-inspection-status",
    type=click.Choice([e.value for e in ContentInspectionStatuses]),
    default=None,
    help="Limit search to sessions with the given content inspection status.",
)
@table_format_option
@columns_option
@logging_options
def search(
    actor_id: Optional[str] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    no_alerts: bool = False,
    risk_indicators: Optional[str] = None,
    state: Optional[List[str]] = None,
    severity: Optional[List[str]] = None,
    rule_id: Optional[List[str]] = None,
    watchlist_id: Optional[List[str]] = None,
    content_inspection_status: Optional[str] = None,
    output: Optional[str] = None,
    certs: Optional[str] = None,
    ignore_cert_validation: Optional[bool] = None,
    checkpoint_name: Optional[str] = None,
    format_: Optional[TableFormat] = None,
    columns: Optional[str] = None,
):
    """
    Search sessions.

    Defaults to only include sessions that have alerts associated with them.
    Use the --no-alerts option to view sessions without any alerts.
    """
    client = Client()
    cursor = _get_cursor_store(client.settings.api_client_id)

    if output:
        format_ = TableFormat.json_lines

    # Use stored checkpoint timestamp for start filter if applicable
    if checkpoint_name:
        checkpoint = cursor.get(checkpoint_name)
        if checkpoint:
            start = int(checkpoint)

    severity_values = []
    for value in severity:
        severity_values.append(SEVERITY_MAP[value])

    sessions_gen = client.sessions.v1.iter_all(
        actor_id=actor_id,
        start_time=start,
        end_time=end,
        has_alerts=not no_alerts,
        risk_indicators=risk_indicators.split(",") if risk_indicators else None,
        states=[*state] if state else None,
        severities=severity_values,
        rule_ids=[*rule_id] if rule_id else None,
        watchlist_ids=[*watchlist_id] if watchlist_id else None,
        content_inspection_status=content_inspection_status,
    )
    if checkpoint_name:
        sessions_gen = _update_checkpoint(cursor, checkpoint_name, sessions_gen)

    with warn_interrupt() if checkpoint_name else nullcontext():
        if output:
            logger = get_server_logger(output, certs, ignore_cert_validation)
            for session in sessions_gen:
                logger.info(session.json())
            return

        if format_ == TableFormat.table:
            columns = columns or [
                "session_id",
                "begin_time",
                "end_time",
                "context_summary",
                "exfiltration_summary",
                "actor_id",
                "low_events",
                "moderate_events",
                "high_events",
                "states",
                "notes",
            ]
            render.table(Session, sessions_gen, columns=columns, flat=False)
        elif format_ == TableFormat.csv:
            render.csv(Session, sessions_gen, columns=columns, flat=True)
        else:
            printed = False
            for session in sessions_gen:
                printed = True
                if format_ == TableFormat.json_pretty:
                    console.print_json(session.json())
                else:
                    click.echo(session.json())
            if not printed:
                console.print("No results found.")


@sessions.command(cls=IncydrCommand)
@click.argument("session-id")
@logging_options
@single_format_option
def show(session_id: str, format_: SingleFormat):
    """
    Show the details of a single session.
    """
    client = Client()
    session = client.sessions.v1.get_session_details(session_id)

    if format_ == SingleFormat.rich:
        console.print(Panel.fit(model_as_card(session)))
    elif format_ == SingleFormat.json_pretty:
        console.print_json(session.json())
    else:
        click.echo(session.json())


@sessions.command(cls=IncydrCommand)
@logging_options
@columns_option
@table_format_option
@output_options
@click.argument("session-id")
def show_events(
    session_id: str,
    columns: Optional[str],
    output: Optional[str],
    certs: Optional[str],
    ignore_cert_validation: Optional[bool],
    format_: SingleFormat,
):
    """
    Show the details of a single session.
    """
    client = Client()
    page = client.sessions.v1.get_session_events(session_id)

    def event_iterator(event_page: FileEventsPage):
        yield from event_page.file_events

    events = event_iterator(page)

    if output:
        logger = get_server_logger(output, certs, ignore_cert_validation)
        for event_ in events:
            logger.info(event_.json())
        return

    if format_ == TableFormat.csv:
        render.csv(FileEventV2, events, columns=columns, flat=True)
    elif format_ == TableFormat.table:
        render.table(FileEventV2, events, columns=columns, flat=False)
    else:
        printed = False
        for event in events:
            printed = True
            if format_ == TableFormat.json_pretty:
                console.print_json(data=event)
            else:
                click.echo(json.dumps(event))
        if not printed:
            console.print("No results found.")


@sessions.command(cls=IncydrCommand)
@click.argument("session-id")
@click.option(
    "--state",
    "-s",
    type=click.Choice([e.value for e in SessionStates]),
    help="Update a session's state.",
)
@click.option(
    "--note",
    help="Add a note to the session. Takes the notes content with a 2000 character max.",
)
@logging_options
def update(session_id: str, state: str = None, note: str = None):
    """
    Update the state of and/or the note attached to the session.
    """
    client = Client()
    if state:
        client.sessions.v1.update_state_by_id(session_id, state)
        console.print(f"Successfully updated session {session_id} to {state}!")
    if note:
        client.sessions.v1.add_note(session_id, note)
        console.print(f"Successfully added note to session {session_id}!")


@sessions.command(cls=IncydrCommand)
@click.argument("file", type=click.File())
@click.option(
    "--state",
    type=click.Choice([e.value for e in SessionStates]),
    help="Override CSV/JSON input's `state` value with this value.",
)
@click.option("--note", help="Override CSV/JSON input's `note` value with this value.")
@input_format_option
@logging_options
def bulk_update_state(
    file: Path,
    state: Optional[str] = None,
    note: Optional[str] = None,
    format_: str = None,
):
    """
    Bulk update the state of multiple sessions. Optionally attach a note.

    NEW_STATE specifies the new state to which sessions will be updated.
    Must be one of the following: 'OPEN', 'IN_PROGRESS', 'CLOSED', 'CLOSED_TP', 'CLOSED_FP', 'OPEN_NEW_DATA'

    Takes a single arg `FILE` which specifies the path to the file (use "-" to read from stdin).
    File format can either be CSV or [JSON Lines format](https://jsonlines.org) (Default is CSV).

    The --state and --note options to this command will override respective columns/keys in the CSV/JSON input.

    Accepts the following columns:

    * `session_id` (required)
    * `state` (required if --state option is not provided)
    * `note` (optional)

    If --state is not provided, the CSV/JSON input _must_ have a `state` column/key for each row/object.

    """
    # if --state is provided, we want that column/key to be optional on input data, otherwise required
    state_type = Optional[str] if state else str

    # Validate input models

    class SessionCSV(CSVModel):
        session_id: str = Field(csv_aliases=["session_id", "sessionId"])
        state: state_type = Field(csv_aliases=["state"])
        note: Optional[str]

    class SessionJSON(Model):
        session_id: str = Field(alias="sessionId")
        state: state_type
        note: Optional[str]

    if format_ == "csv":
        models = SessionCSV.parse_csv(file)
    else:
        models = SessionJSON.parse_json_lines(file)

    # Process input

    client = Client()

    # optimization: bucketize update state calls in batches of 100

    try:
        for row in track(
            models,
            description="Updating sessions...",
            transient=True,
        ):
            client.sessions.v1.update_state_by_id(row.session_id, state or row.state)
            if note or row.note:
                client.sessions.v1.add_note(row.session_id, note or row.note)
    except HTTPError as err:
        console.print(f"[red]Error:[/red] {err.response.text}")


def _get_cursor_store(api_key):
    """
    Get cursor store for sessions search checkpoints.
    """
    dir_path = get_user_project_path(
        "checkpoints",
        api_key,
        "session_checkpoints",
    )
    return CursorStore(dir_path, "sessions")


def _update_checkpoint(cursor, checkpoint_name, sessions_gen):
    """
    De-duplicates results across checkpointed runs.

    Since using the timestamp of the last result
    processed as the `--start` time of the next run causes the last event to show up again in the
    next results, store the last session IDs in the cursor to
    filter out on the next run.

    It's also possible that two sessions have the exact same timestamp, so
    `checkpoint_sessions` needs to be a list of session IDs so we can filter out everything that's actually
    been processed.
    """
    checkpoint_sessions = cursor.get_items(checkpoint_name)
    new_timestamp = None
    new_sessions = []
    for session in sessions_gen:
        session_id = session.session_id
        if session_id not in checkpoint_sessions:
            if not new_timestamp or session.end_time > new_timestamp:
                new_timestamp = session.end_time
                new_sessions.clear()
            new_sessions.append(session_id)
            yield session
            cursor.replace(checkpoint_name, new_timestamp)
            cursor.replace_items(checkpoint_name, new_sessions)



from __future__ import annotations

from itertools import chain
from itertools import repeat
from typing import Any
from typing import Generator
from typing import List
from typing import Tuple
from typing import Type

import rich.box
from pydantic import BaseModel
from pydantic.fields import ModelField
from pydantic.fields import SHAPE_SINGLETON
from rich.console import ConsoleRenderable
from rich.console import Group
from rich.console import group
from rich.panel import Panel
from rich.text import Text


def get_field_value_and_info(
    model: BaseModel, path: List[str]
) -> Tuple[Any, ModelField]:
    """
    Traverse a pydantic model and its sub-models to retrieve both the value and `ModelField` data for a given attribute
    path.

    For example, given the following model hierarchy:

        class Child(BaseModel):
            field_1: str = Field("example", extra_data=1)
            field_2: int = Field(1)

        class Parent(BaseModel):
            field: str = Field("default")
            child: Child


    >>> model = Parent(child=Child())
    >>> value, field = get_field_value_and_info(model, path=["child", "field_1"])

    The `value` var would contain the string "example", and `field` would be the Field object for `Child.field_1`, where
    the 'extra_data' would be accessible in `field.field_info.extra`.
    """
    for p in path[:-1]:
        next_model = getattr(model, p)
        # if a child model is Optional and _not_ present, we can inspect the parent field data to get the missing child
        # class and use that to "construct" an empty version of the model, so child fields will still return valid
        # field_info, but the value will be `None` for every field on "missing" child models.
        if next_model is None:
            model_type = model.__fields__[p].type_
            model = model_type.construct(
                **{field: None for field in model_type.__fields__}
            )
        else:
            model = next_model
    value = getattr(model, path[-1])
    field = model.__fields__.get(path[-1])
    return value, field


def iter_model_formatted(
    model: BaseModel,
    include: List[str] = None,
    flat: bool = False,
    render: str = None,
):
    """
    Iterates through the fields of a Model (with optional flattening of sub-models), yielding (name, value) pairs.

    Accepts a list of field names to filter by (if flat=True, `include` list must be flattened dot-notation names).

    Will automatically attempt to "render" the field values in the following order:
       - if `render` arg is a string, it will look in the "extra" section of the pydantic model's Field Info for that
          name (expects a callable to be there)
       - if value is of a type that the model has a `json_encoder` for, it will use that encoder
       - otherwise will leave the value unchanged
    """
    fields = get_fields(model.__class__, include=include, flat=flat)
    for name in fields:
        path = name.split(".")
        value, field = get_field_value_and_info(model, path)
        field_renderer = None if not field else field.field_info.extra.get(render)
        if render and field_renderer:
            value = field_renderer(value)
            yield name, value
            continue
        json_encoder = model.Config.json_encoders.get(type(value))
        if json_encoder:
            value = json_encoder(value)
        yield name, value


def list_as_panel(
    items, sep=None, title=None, expand=True, box=rich.box.ROUNDED
) -> Panel:
    """
    Renders a list of items as a `rich.panel.Panel`. If `sep` is provided will add a separator between each item (useful
    for rendering list of models that have multiple values.

    Examples:
        >>> from rich import print
        >>> print(list_as_panel(["one", "two", "three", "four"]))
        ╭───────╮
        │ one   │
        │ two   │
        │ three │
        │ four  │
        ╰───────╯

        >>> print(list_as_panel(["one", "two", "three", "four"], sep="---", title="example"))
        ╭─ example ─╮
        │ one       │
        │ ---       │
        │ two       │
        │ ---       │
        │ three     │
        │ ---       │
        │ four      │
        ╰───────────╯
    """
    if sep:
        # zip has an added `strict=False` kwarg as of python 3.10
        # flake8's B905 error will call this out, suppressed for now for backwards compatibility with <3.10
        items = list(chain(*zip(items, repeat(sep))))[:-1]
    return Panel(Group(*items), title=title, expand=expand, box=box)


@group()
def model_as_card(model, include=None):
    """
    Renders a pydantic model in 'card' format, where field name/value pairs are presented vertically, and when a field
    is a list of items, it renders it as a separate panel.
    """
    for name, value in iter_model_formatted(
        model, include=include, flat=True, render="table"
    ):
        # rendering list fields takes some special handling
        if isinstance(value, list):
            if not len(value):
                yield f"{name}: []"
            # since a list of models can contain many values for each "item" in the list, we want to wrap the models
            # with a separator to make it readable and easily distinguish where each model begins/ends.
            elif isinstance(value[0], BaseModel):
                yield list_as_panel(
                    [model_as_card(v) for v in value], title=name, sep="---"
                )
            else:
                yield list_as_panel(value, title=name)
        # this should only be true when a model field has a "table" extra renderer defined
        elif isinstance(value, ConsoleRenderable):
            yield Panel.fit(value, title=name)
        else:
            yield Text(f"{name}: {value}", overflow="fold")


def flatten_fields(model: Type[BaseModel]) -> Generator[str, None, None]:
    """
    Yields all fields of a model and any sub-models as flat, dot-notation strings.

    For example, given the following model hierarchy:

        class Child(BaseModel):
            field_1: str
            field_2: int

        class Parent(BaseModel):
            field: str
            child: Child

    flatten_fields(Parent) would yield: ['field', 'child.field_1', 'child.field_2']
    """
    for name, field in model.__fields__.items():
        # the field.shape tells us if the field contains a single `BaseModel` or something like a `List[BaseModel]`
        # we can only traverse singleton models when flattening
        try:
            is_subclass = issubclass(field.type_, BaseModel)
        # TypeError is thrown if field.type_ is a Union type.
        # Assumes our endpoints won't return a field that can be one of multiple models
        # This would be a pretty odd API design anyway
        except TypeError:
            is_subclass = False
        if field.shape == SHAPE_SINGLETON and is_subclass:
            for child_name in flatten_fields(field.type_):
                yield f"{name}.{child_name}"
        else:
            yield name


def get_fields(
    model: Type[BaseModel], include: List[str] = None, flat: bool = False
) -> Generator[str, None, None]:
    """
    Yields fields from a model.

    If flat=True will flatten nested models.

    Results can be filtered with `include` param list. Values can either be the field name, or a wildcard name which
    will include subfields matching that name pattern. For example: ["event.id", "file.*"] will yield the event.id
    field and _all_ file fields of the nested FileEventV2 model.

    Order is preserved to match `include` order, to allow precise table/csv column ordering based on user input.
    """
    fields = list(flatten_fields(model)) if flat else model.__fields__
    if not include:
        yield from fields
    else:
        for i in include:
            found = False
            if "*" in i:
                pattern = i.replace("*", "")
                for f in fields:
                    if f.startswith(pattern):
                        found = True
                        yield f
            elif i in fields:
                found = True
                yield i
            if not found:
                raise ValueError(
                    f"'{i}' is not a valid field path for model: {model.__name__}",
                    list(fields),
                )



from __future__ import annotations

import json
from datetime import datetime
from typing import List
from typing import Optional

import rich.box
from pydantic import constr
from pydantic import Field
from pydantic import validator
from rich.markdown import Markdown
from rich.panel import Panel

from _incydr_sdk.core.models import Model


class Observation(Model):
    id: Optional[str] = Field(
        None, description="Id of given observation.", example="uniqueObservationId"
    )
    observed_at: datetime = Field(
        ...,
        alias="observedAt",
        description="Timestamp when the activity was first observed.",
        example="2020-02-19T01:57:45.006683Z",
    )
    last_observed_at: Optional[datetime] = Field(
        None,
        alias="lastObservedAt",
        description="Timestamp when the activity was last observed.",
        example="2020-02-19T01:57:45.006683Z",
    )
    type: Optional[str] = Field(
        None,
        description="The type of observation data recorded.",
        example="FedCloudSharePermissions",
    )
    data: Optional[dict] = Field(
        None,
        description="The JSON formatted observation data rolled into one aggregation.",
        example='{"type$":"OBSERVED_CLOUD_SHARE_ACTIVITY","id":"exampleId","sources":["OneDrive"],"exposureTypes":["PublicLinkShare"],"firstActivityAt":"2020-02-19T01:50:00.0000000Z","lastActivityAt":"2020-02-19T01:55:00.0000000Z","fileCount":2,"totalFileSize":200,"fileCategories":[{"type$":"OBSERVED_FILE_CATEGORY","category":"Document","fileCount":2,"totalFileSize":53,"isSignificant":false}],"outsideTrustedDomainsEmailsCount":0,"outsideTrustedDomainsTotalDomainCount":0,"outsideTrustedDomainsTotalDomainCountTruncated":false}',
        table=lambda data: json.dumps(data, indent=2),
    )

    @validator("data", pre=True)
    def parse_data_json(cls, value):  # noqa
        return json.loads(value) if value else None


class Note(Model):
    id: Optional[str] = Field(
        None, description="Unique id of the note.", example="noteId"
    )
    last_modified_at: datetime = Field(
        ...,
        alias="lastModifiedAt",
        description="Timestamp of when the note was last modified.",
        example="2020-02-19T01:57:45.006683Z",
    )
    last_modified_by: Optional[str] = Field(
        None,
        alias="lastModifiedBy",
        description="User who last modified the note.",
        example="exampleUsername",
    )
    message: Optional[str] = Field(
        None,
        description="The note itself.",
        example="This is a note.",
        table=lambda note: Panel(Markdown(note), width=120, box=rich.box.MINIMAL)
        if note
        else note,
    )


class AuditInfo(Model):
    modified_by: Optional[str] = Field(
        None,
        alias="modifiedBy",
        description="Username of the individual who last modified the rule.",
        example="UserWhoMostRecentlyModifiedTheRule",
    )
    modified_at: datetime = Field(
        ...,
        alias="modifiedAt",
        description="Timestamp of when the rule was last modified.",
        example="2020-02-19T01:57:45.006683Z",
    )


class Watchlist(Model):
    id: Optional[str] = Field(
        None, description="Unique id of this watchlist.", example="guid"
    )
    name: Optional[str] = Field(
        None, description="Name of the watchlist.", example="Development Department"
    )
    type: Optional[str] = Field(
        ..., description="Type of watchlist.", example="DEPARTING_EMPLOYEE"
    )
    is_significant: bool = Field(
        ...,
        alias="isSignificant",
        description="Indicates whether the watchlist was part of the triggering rule's criteria.",
        example="true",
    )


class ObserverRuleMetadata(AuditInfo):
    name: Optional[str] = Field(
        None,
        description="The name of the rule.",
        example="My Removable Media Exfiltration Rule",
    )
    description: Optional[str] = Field(
        None,
        description="The description of the rule.",
        example="Will generate alerts when files moved to USB.",
    )
    severity: Optional[str] = Field(
        None, description="The static severity of the rule (deprecated)."
    )
    is_system: Optional[bool] = Field(
        None,
        alias="isSystem",
        description="Boolean indicating if the rule was created from another Code42 Application.",
        example="FALSE",
    )
    is_enabled: bool = Field(
        ...,
        alias="isEnabled",
        description="Boolean indicating if the rule is enabled to trigger alerts.",
        example="TRUE",
    )
    rule_source: Optional[str] = Field(
        None,
        alias="ruleSource",
        description="The source of the rule.  Will be one of [DepartingEmployee, Alerting, HighRiskEmployee]",
        example="Alerting",
    )


class AlertSummary(Model):
    """
    A model representing an alert summary.

    **Fields**:

    * **tenant_id**: `str` The unique identifier representing the tenant.
    * **type**: `RuleType` Rule type that generated the alert.
    * **id**: `str` The unique id of the alert.
    * **created_at**: `datetime` The timestamp when the alert was created.
    * **state**: [`AlertState`][alert-state] The current state of the alert.
    * **state_last_modified_by**: `str` The actor who last modified the alert state.
    * **state_last_modified_at**: `datetime` The timestamp when the alert state was last modified.
    * **name**: `str` The name of the alert.  Same as the name of the rule that triggered it.
    * **description**: `str` The description of the alert.  Same as the description of the rule that triggered it.
    * **actor**: `str` The user who triggered the alert.
    * **actor_id**: `str` The user id who triggered the alert, if it is available.
    * **target**: `str` Unused legacy property.
    * **severity**: [`AlertSeverity`][alert-severity] Indicates static rule severity of the alert. (Deprecated)
    * **risk_severity**: [`RiskSeverity`][risk-severity] Indicates event risk severity of the alert.
    * **rule_id**: `str` The unique id corresponding to the rule which triggered the alert.
    * **watchlists**: `str` Watchlists the actor is on at the time of the alert (if any).
    """

    tenant_id: constr(max_length=40) = Field(
        ...,
        alias="tenantId",
        description="The unique identifier representing the tenant.",
        example="MyExampleTenant",
    )
    type: Optional[str] = Field(..., description="Rule type that generated the alert.")
    id: Optional[str] = Field(
        None, description="The unique id of the alert.", example="alertId"
    )
    created_at: datetime = Field(
        ...,
        alias="createdAt",
        description="The timestamp when the alert was created.",
        example="2020-02-19T01:57:45.006683Z",
    )
    state: Optional[str] = Field(..., description="The current state of the alert.")
    state_last_modified_by: Optional[str] = Field(None, alias="stateLastModifiedBy")
    state_last_modified_at: Optional[datetime] = Field(
        None, alias="stateLastModifiedAt"
    )
    name: Optional[str] = Field(
        None,
        description="The name of the alert.  Same as the name of the rule that triggered it.",
        example="Removable Media Exfiltration Rule",
    )
    description: Optional[str] = Field(
        None,
        description="The description of the alert.  Same as the description of the rule that triggered it.",
        example="Alert me on all removable media exfiltration.",
    )
    actor: Optional[str] = Field(
        None,
        description="The user who triggered the alert.",
        example="exampleUser@mycompany.com",
    )
    actor_id: Optional[str] = Field(
        None,
        alias="actorId",
        description="The authority user id who triggered the alert, if it is available.",
        example="authorityUserId",
    )
    target: Optional[str] = None
    severity: Optional[str] = Field(
        None, description="Indicates static rule severity of the alert."
    )
    risk_severity: Optional[str] = Field(
        None,
        alias="riskSeverity",
        description="Indicates event risk severity of the alert.",
        example="MODERATE",
    )
    rule_id: Optional[str] = Field(
        None,
        alias="ruleId",
        description="The unique id corresponding to the rule which triggered the alert.",
        example="uniqueRuleId",
    )
    watchlists: Optional[List[Watchlist]] = Field(
        None,
        description="Watchlists the actor is on at the time of the alert.",
        example=[],
        table=lambda watchlists: "\n".join((w.name or w.type) for w in watchlists)
        if watchlists
        else watchlists,
        csv=lambda watchlists: "|".join((w.name or w.type) for w in watchlists)
        if watchlists
        else watchlists,
    )


class ObserverRuleMetadataEssentials(ObserverRuleMetadata):
    tenant_id: constr(max_length=40) = Field(
        ...,
        alias="tenantId",
        description="The unique identifier representing the tenant.",
        example="MyExampleTenant",
    )
    observer_rule_id: Optional[str] = Field(
        None,
        alias="observerRuleId",
        description="Id of the rule in the observer.",
        example="UniqueRuleId",
    )
    type: Optional[str] = Field(..., description="Rule type of the rule.")


class AlertDetails(AlertSummary):
    """
    A model representing the full details of an alert. Includes all the fields from `AlertSummary` plus file event
    observations (the events that triggered the alert), and any notes that have been added to the alert.

    **Fields**:

    * **observations**: `List[Observation]` List of observed file events that triggered the alert.
    * **note**: `str` Most recent note added to the alert.
    """

    observations: Optional[List[Observation]]
    note: Optional[Note]



from __future__ import annotations

from typing import List
from typing import Optional

from pydantic import BaseModel
from pydantic import constr
from pydantic import Field

from _incydr_sdk.enums.alerts import AlertState


class UpdateAlertStateRequest(BaseModel):
    tenant_id: constr(max_length=40) = Field(
        ...,
        alias="tenantId",
        description="The unique identifier representing the tenant.",
        example="MyExampleTenant",
    )
    alert_ids: List[str] = Field(
        ...,
        alias="alertIds",
        description="The unique identifiers representing the alerts you want to act upon.",
        example=["ExampleAlertId1", "ExampleAlertId2"],
        max_length=100,
    )
    state: AlertState = Field(
        ..., description="The state to update the given alerts to."
    )
    note: Optional[constr(max_length=2000)] = Field(
        None,
        description="An optional note to attach to the alert",
        example="This is an example note.",
    )


class AlertDetailsRequest(BaseModel):
    alert_ids: List[str] = Field(
        ...,
        alias="alertIds",
        description="The unique identifiers representing the alerts you want to act upon.",
        example=["ExampleAlertId1", "ExampleAlertId2"],
        max_length=100,
    )


class AddNoteRequest(BaseModel):
    tenant_id: constr(max_length=40) = Field(
        ...,
        alias="tenantId",
        description="The unique identifier representing the tenant.",
        example="MyExampleTenant",
    )
    alert_id: constr(max_length=40) = Field(
        ...,
        alias="alertId",
        description="The unique identifier representing the alert you want to act upon.",
        example="ExampleAlertId",
    )
    note: constr(max_length=2000) = Field(
        ...,
        description="The note to attach to the alert.",
        example="This is an example note.",
    )



from _incydr_sdk.enums import _Enum


class Operator(_Enum):
    # all valid filter operators for querying file events
    IS = "IS"
    IS_NOT = "IS_NOT"
    EXISTS = "EXISTS"
    DOES_NOT_EXIST = "DOES_NOT_EXIST"
    GREATER_THAN = "GREATER_THAN"
    LESS_THAN = "LESS_THAN"
    ON = "ON"
    ON_OR_AFTER = "ON_OR_AFTER"
    ON_OR_BEFORE = "ON_OR_BEFORE"
    WITHIN_THE_LAST = "WITHIN_THE_LAST"


class EventSearchTerm(_Enum):
    """
    Search terms available to filtering file events.
    """

    TIMESTAMP = "@timestamp"
    DESTINATION_ACCOUNT_NAME = "destination.accountName"
    DESTINATION_ACCOUNT_TYPE = "destination.accountType"
    DESTINATION_CATEGORY = "destination.category"
    DESTINATION_DOMAINS = "destination.domains"
    DESTINATION_EMAIL_RECIPIENTS = "destination.email.recipients"
    DESTINATION_EMAIL_SUBJECT = "destination.email.subject"
    DESTINATION_IP = "destination.ip"
    DESTINATION_NAME = "destination.name"
    DESTINATION_OPERATING_SYSTEM = "destination.operatingSystem"
    DESTINATION_PRINT_JOB_NAME = "destination.printJobName"
    DESTINATION_PRINTED_FILES_BACKUP_PATH = "destination.printedFilesBackupPath"
    DESTINATION_PRINTER_NAME = "destination.printerName"
    DESTINATION_PRIVATE_IP = "destination.privateIp"
    DESTINATION_REMOTE_HOSTNAME = "destination.remoteHostname"
    DESTINATION_REMOVABLE_MEDIA_BUS_TYPE = "destination.removableMedia.busType"
    DESTINATION_REMOVABLE_MEDIA_CAPACITY = "destination.removableMedia.capacity"
    DESTINATION_REMOVABLE_MEDIA_MEDIA_NAME = "destination.removableMedia.mediaName"
    DESTINATION_REMOVABLE_MEDIA_NAME = "destination.removableMedia.name"
    DESTINATION_REMOVABLE_MEDIA_PARTITION_ID = "destination.removableMedia.partitionId"
    DESTINATION_REMOVABLE_MEDIA_SERIAL_NUMBER = (
        "destination.removableMedia.serialNumber"
    )
    DESTINATION_REMOVABLE_MEDIA_VENDOR = "destination.removableMedia.vendor"
    DESTINATION_REMOVABLE_MEDIA_VOLUME_NAME = "destination.removableMedia.volumeName"
    DESTINATION_TABS_TITLE = "destination.tabs.title"
    DESTINATION_TABS_TITLE_ERROR = "destination.tabs.titleError"
    DESTINATION_TABS_URL = "destination.tabs.url"
    DESTINATION_TABS_URL_ERROR = "destination.tabs.urlError"
    DESTINATION_USER_EMAIL = "destination.user.email"
    EVENT_ACTION = "event.action"
    EVENT_ID = "event.id"
    EVENT_INGESTED = "event.ingested"
    EVENT_INSERTED = "event.inserted"
    EVENT_OBSERVER = "event.observer"
    EVENT_RELATED_EVENTS_AGENT_TIMESTAMP = "event.relatedEvents.agentTimestamp"
    EVENT_RELATED_EVENTS_EVENT_ACTION = "event.relatedEvents.eventAction"
    EVENT_RELATED_EVENTS_ID = "event.relatedEvents.id"
    EVENT_RELATED_EVENTS_SOURCE_CATEGORY = "event.relatedEvents.sourceCategory"
    EVENT_RELATED_EVENTS_SOURCE_NAME = "event.relatedEvents.sourceName"
    EVENT_RELATED_EVENTS_TABS_TITLE = "event.relatedEvents.tabs.title"
    EVENT_RELATED_EVENTS_TABS_TITLE_ERROR = "event.relatedEvents.tabs.titleError"
    EVENT_RELATED_EVENTS_TABS_URL = "event.relatedEvents.tabs.url"
    EVENT_RELATED_EVENTS_TABS_URL_ERROR = "event.relatedEvents.tabs.urlError"
    EVENT_RELATED_EVENTS_USER_EMAIL = "event.relatedEvents.userEmail"
    EVENT_SHARE_TYPE = "event.shareType"
    EVENT_DETECTOR_DISPLAY_NAME = "event.detectorDisplayName"
    EVENT_VECTOR = "event.vector"
    FILE_ACQUIRED_FROM_GIT_REPOSITORY_EMAIL = "file.acquiredFrom.git.repositoryEmail"
    FILE_ACQUIRED_FROM_GIT_REPOSITORY_URI = "file.acquiredFrom.git.repositoryUri"
    FILE_ACQUIRED_FROM_GIT_REPOSITORY_USER = "file.acquiredFrom.git.repositoryUser"
    FILE_ACQUIRED_FROM_TABS_URL_ERROR = "file.acquiredFrom.tabs.urlError"
    FILE_ACQUIRED_FROM_SOURCE_DOMAINS = "file.acquiredFrom.sourceDomains"
    FILE_ACQUIRED_FROM_USER_EMAIL = "file.acquiredFrom.userEmail"
    FILE_ACQUIRED_FROM_SOURCE_USER_EMAIL = "file.acquiredFrom.sourceUser.email"
    FILE_ACQUIRED_FROM_TABS_URL = "file.acquiredFrom.tabs.url"
    FILE_ACQUIRED_FROM_EVENT_ID = "file.acquiredFrom.eventId"
    FILE_ACQUIRED_FROM_FILE_NAME = "file.acquiredFrom.fileName"
    FILE_ACQUIRED_FROM_SOURCE_CATEGORY = "file.acquiredFrom.sourceCategory"
    FILE_ACQUIRED_FROM_SOURCE_ACCOUNT_NAME = "file.acquiredFrom.sourceAccountName"
    FILE_ACQUIRED_FROM_SOURCE_NAME = "file.acquiredFrom.sourceName"
    FILE_ACQUIRED_FROM_AGENT_TIMESTAMP = "file.acquiredFrom.agentTimestamp"
    FILE_ACQUIRED_FROM_EVENT_ACTION = "file.acquiredFrom.eventAction"
    FILE_ACQUIRED_FROM_TABS_TITLE_ERROR = "file.acquiredFrom.tabs.titleError"
    FILE_ACQUIRED_FROM_TABS_TITLE = "file.acquiredFrom.tabs.title"
    FILE_ACQUIRED_FROM_SOURCE_ACCOUNT_TYPE = "file.acquiredFrom.sourceAccountType"
    FILE_ACQUIRED_FROM_MD5 = "file.acquiredFrom.md5"
    FILE_CATEGORY = "file.category"
    FILE_CATEGORY_BY_BYTES = "file.categoryByBytes"
    FILE_CATEGORY_BY_EXTENSION = "file.categoryByExtension"
    FILE_CHANGE_TYPE = "file.changeType"
    FILE_CLASSIFICATIONS_VALUE = "file.classifications.value"
    FILE_CLASSIFICATIONS_VENDOR = "file.classifications.vendor"
    FILE_CLOUD_DRIVE_ID = "file.cloudDriveId"
    FILE_CREATED = "file.created"
    FILE_DIRECTORY = "file.directory"
    FILE_DIRECTORY_ID = "file.directoryId"
    FILE_HASH_MD5 = "file.hash.md5"
    FILE_HASH_MD5_ERROR = "file.hash.md5Error"
    FILE_HASH_SHA256 = "file.hash.sha256"
    FILE_HASH_SHA256_ERROR = "file.hash.sha256Error"
    FILE_ID = "file.id"
    FILE_MIME_TYPE_BY_BYTES = "file.mimeTypeByBytes"
    FILE_MIME_TYPE_BY_EXTENSION = "file.mimeTypeByExtension"
    FILE_MODIFIED = "file.modified"
    FILE_NAME = "file.name"
    FILE_ORIGINAL_NAME = "file.originalName"
    FILE_ORIGINAL_DIRECTORY = "file.originalDirectory"
    FILE_OWNER = "file.owner"
    FILE_SIZE_IN_BYTES = "file.sizeInBytes"
    FILE_URL = "file.url"
    GIT_REPOSITORY_URI = "git.repositoryUri"
    GIT_LAST_COMMIT_HASH = "git.lastCommitHash"
    GIT_EVENT_ID = "git.eventId"
    GIT_REPOSITORY_ENDPOINT_PATH = "git.repositoryEndpointPath"
    GIT_REPOSITORY_USER = "git.repositoryUser"
    GIT_REPOSITORY_EMAIL = "git.repositoryEmail"
    PROCESS_EXECUTABLE = "process.executable"
    PROCESS_OWNER = "process.owner"
    REPORT_COUNT = "report.count"
    REPORT_DESCRIPTION = "report.description"
    REPORT_HEADERS = "report.headers"
    REPORT_ID = "report.id"
    REPORT_NAME = "report.name"
    REPORT_TYPE = "report.type"
    RISK_INDICATORS_NAME = "risk.indicators.name"
    RISK_INDICATORS_WEIGHT = "risk.indicators.weight"
    RISK_INDICATORS_ID = "risk.indicators.id"
    RISK_SCORE = "risk.score"
    RISK_SEVERITY = "risk.severity"
    RISK_TRUST_REASON = "risk.trustReason"
    RISK_TRUSTED = "risk.trusted"
    RISK_UNTRUSTED_VALUES_GIT_REPOSITORY_URIS = "risk.untrustedValues.gitRepositoryUris"
    RISK_UNTRUSTED_VALUES_URL_PATHS = "risk.untrustedValues.urlPaths"
    RISK_UNTRUSTED_VALUES_DOMAINS = "risk.untrustedValues.domains"
    RISK_UNTRUSTED_VALUES_SLACK_WORKSPACES = "risk.untrustedValues.slackWorkspaces"
    RISK_UNTRUSTED_VALUES_ACCOUNT_NAMES = "risk.untrustedValues.accountNames"
    SOURCE_ACCOUNT_TYPE = "source.accountType"
    SOURCE_ACCOUNT_NAME = "source.accountName"
    SOURCE_CATEGORY = "source.category"
    SOURCE_DOMAIN = "source.domain"
    SOURCE_DOMAINS = "source.domains"
    SOURCE_EMAIL_FROM = "source.email.from"
    SOURCE_EMAIL_SENDER = "source.email.sender"
    SOURCE_IP = "source.ip"
    SOURCE_NAME = "source.name"
    SOURCE_OPERATING_SYSTEM = "source.operatingSystem"
    SOURCE_PRIVATE_IP = "source.privateIp"
    SOURCE_REMOVABLE_MEDIA_BUS_TYPE = "source.removableMedia.busType"
    SOURCE_REMOVABLE_MEDIA_CAPACITY = "source.removableMedia.capacity"
    SOURCE_REMOVABLE_MEDIA_MEDIA_NAME = "source.removableMedia.mediaName"
    SOURCE_REMOVABLE_MEDIA_NAME = "source.removableMedia.name"
    SOURCE_REMOVABLE_MEDIA_PARTITION_ID = "source.removableMedia.partitionId"
    SOURCE_REMOVABLE_MEDIA_SERIAL_NUMBER = "source.removableMedia.serialNumber"
    SOURCE_REMOVABLE_MEDIA_VENDOR = "source.removableMedia.vendor"
    SOURCE_REMOVABLE_MEDIA_VOLUME_NAME = "source.removableMedia.volumeName"
    SOURCE_TABS_TITLE = "source.tabs.title"
    SOURCE_TABS_TITLE_ERROR = "source.tabs.titleError"
    SOURCE_TABS_URL = "source.tabs.url"
    SOURCE_TABS_URL_ERROR = "source.tabs.urlError"
    SOURCE_USER_EMAIL = "source.user.email"
    USER_DEVICE_UID = "user.deviceUid"
    USER_EMAIL = "user.email"
    USER_ID = "user.id"


class FileCategory(_Enum):
    """Available file categories for filtering file events."""

    AUDIO = "Audio"
    DOCUMENT = "Document"
    EXECUTABLE = "Executable"
    IMAGE = "Image"
    PDF = "Pdf"
    PRESENTATION = "Presentation"
    SCRIPT = "Script"
    SOURCE_CODE = "SourceCode"
    SPREADSHEET = "Spreadsheet"
    VIDEO = "Video"
    VIRTUAL_DISK_IMAGE = "VirtualDiskImage"
    ZIP = "Archive"


class EventAction(_Enum):
    """Available event actions for filtering file events."""

    # Exposure Type in v1
    REMOVABLE_MEDIA_CREATED = "removable-media-created"
    REMOVABLE_MEDIA_MODIFIED = "removable-media-modified"
    REMOVABLE_MEDIA_DELETED = "removable-media-deleted"
    SYNC_APP_CREATED = "sync-app-created"
    SYNC_APP_MODIFIED = "sync-app-modified"
    SYNC_APP_DELETED = "sync-app-deleted"
    FILE_SHARED = "file-shared"

    # Event Type in v1
    FILE_CREATED = "file-created"
    FILE_DELETED = "file-deleted"
    FILE_DOWNLOADED = "file-downloaded"
    FILE_EMAILED = "file-emailed"
    FILE_MODIFIED = "file-modified"
    FILE_PRINTED = "file-printed"
    APPLICATION_READ = "application-read"


class Category(_Enum):
    """Source and destination categories available for filtering file events."""

    BUSINESS_TOOLS = "Business Tools"
    CLOUD_STORAGE = "Cloud Storage"
    DEVICE = "Device"
    EMAIL = "Email"
    MESSAGING = "Messaging"
    MULTIPLE_POSSIBILITIES = "Multiple Possibilities"
    SOCIAL_MEDIA = "Social Media"
    SOURCE_CODE_REPOSITORY = "Source Code Repository"
    UNCATEGORIZED = "Uncategorized"
    UNKNOWN = "Unknown"
    BUSINESS_INTELLIGENCE_TOOLS = "Business Intelligence Tools"
    CIVIL_SERVICES = "Civil Services"
    CLOUD_COMPUTING = "Cloud Computing"
    CODING_TOOLS = "Coding Tools"
    CONTRACT_MANAGEMENT = "Contract Management"
    CRM_TOOLS = "CRM Tools"
    DESIGN_TOOLS = "Design Tools"
    E_COMMERCE = "E-commerce"
    FILE_CONVERSION_TOOLS = "File Conversion Tools"
    FINANCIAL_SERVICES = "Financial Services"
    HEALTHCARE_AND_INSURANCE = "Healthcare & Insurance"
    HR_TOOLS = "HR Tools"
    IMAGE_HOSTING = "Image Hosting"
    IT_SERVICES = "IT Services"
    JOB_LISTINGS = "Job Listings"
    LEARNING_PLATFORMS = "Learning Platforms"
    MARKETING_TOOLS = "Marketing Tools"
    PDF_MANAGER = "PDF Manager"
    PHOTO_PRINTING = "Photo Printing"
    PRODUCTIVITY_TOOLS = "Productivity Tools"
    PROFESSIONAL_SERVICES = "Professional Services"
    REAL_ESTATE = "Real Estate"
    SALES_TOOLS = "Sales Tools"
    SEARCH_ENGINE = "Search Engine"
    SHIPPING = "Shipping"
    SOFTWARE = "Software"
    TRAVEL = "Travel"
    WEB_HOSTING = "Web Hosting"


class ShareType(_Enum):
    """Share types available for filtering file events."""

    PUBLIC_LINK_SHARE = "Anyone with the link"
    DOMAIN_SHARE = "Anyone in your organization"
    DIRECT_USER_SHARE = "Shared with specific people"


class ReportType(_Enum):
    """Report types available for filtering file events."""

    AD_HOC = "REPORT_TYPE_AD_HOC"
    SAVED = "REPORT_TYPE_SAVED"


class RiskIndicators(_Enum):
    """Risk indicator names available for filtering file events."""

    # Destinations
    ADOBE_UPLOAD = "Adobe upload"
    ADOBE_ACROBAT_UPLOAD = "Adobe Acrobat upload"
    AIR_DROP = "AirDrop"
    AMAZON_DRIVE_UPLOAD = "Amazon Drive upload"
    AOL_UPLOAD = "AOL upload"
    BAIDU_NET_DISK_UPLOAD = "Baidu NetDisk upload"
    BITBUCKET_UPLOAD = "Bitbucket upload"
    BOX_UPLOAD = "Box upload"
    CANVA_UPLOAD = "Canva upload"
    CLOUD_CONVERT_UPLOAD = "CloudConvert upload"
    COLABORATORY_UPLOAD = "Colaboratory upload"
    COMBINE_PDF_UPLOAD = "CombinePDF upload"
    COMCAST_UPLOAD = "Comcast upload"
    COMPRESS_JPEG_UPLOAD = "Compress JPEG upload"
    CRASHPLAN_UPLOAD = "Crashplan upload"
    DISCORD_UPLOAD = "Discord upload"
    DRAKE_PORTALS_UPLOAD = "Drake Portals upload"
    DROPBOX_UPLOAD = "Dropbox upload"
    EVERNOTE_UPLOAD = "Evernote upload"
    FACEBOOK_MESSENGER_UPLOAD = "Facebook Messenger upload"
    FACEBOOK_UPLOAD = "Facebook upload"
    FASTMAIL_UPLOAD = "Fastmail upload"
    FIGMA_UPLOAD = "Figma upload"
    FILE_DOT_IO_UPLOAD = "File.io upload"
    FILESTACK_UPLOAD = "Filestack upload"
    FOUR_CHAN_UPLOAD = "4chan upload"
    FREE_CONVERT_UPLOAD = "Free Convert upload"
    FREE_PDF_CONVERT_UPLOAD = "Free PDF Convert upload"
    GIT_HUB_UPLOAD = "GitHub upload"
    GIT_HUB_PAGES_UPLOAD = "GitHub Pages upload"
    GIT_LAB_UPLOAD = "GitLab upload"
    GMAIL_UPLOAD = "Gmail upload"
    GMX_UPLOAD = "GMX upload"
    GOOGLE_APPS_SCRIPT_UPLOAD = "Google Apps Script upload"
    GOOGLE_CHAT_UPLOAD = "Google Chat upload"
    GOOGLE_CLOUD_SHELL_UPLOAD = "Google Cloud Shell upload"
    GOOGLE_DRIVE_UPLOAD = "Google Drive upload"
    GOOGLE_HANGOUTS_UPLOAD = "Google Hangouts upload"
    GOOGLE_JAMBOARD_UPLOAD = "Google Jamboard upload"
    GOOGLE_KEEP_UPLOAD = "Google Keep upload"
    GOOGLE_MESSAGES_UPLOAD = "Google Messages upload"
    GOOGLE_SITES_UPLOAD = "Google Sites upload"
    HEIC_TO_JPEG_UPLOAD = "HEICtoJPEG upload"
    ICLOUD_MAIL_UPLOAD = "iCloud Mail upload"
    ICLOUD_UPLOAD = "iCloud upload"
    I_LOVE_PDF_UPLOAD = "iLovePDF upload"
    IMAGE_COLOR_PICKER_UPLOAD = "Image Color Picker upload"
    IMGUR_UPLOAD = "Imgur upload"
    JPG2_PDF_UPLOAD = "JPG2PDF upload"
    KAPWING_UPLOAD = "Kapwing upload"
    LINKED_IN_UPLOAD = "LinkedIn upload"
    LYCOS_UPLOAD = "Lycos upload"
    MAIL_COM_UPLOAD = "Mail.com upload"
    MEGA_UPLOAD = "Mega upload"
    MICROSOFT_TEAMS_UPLOAD = "Microsoft Teams upload"
    MIRO_UPLOAD = "Miro upload"
    MONDAY_UPLOAD = "Monday upload"
    MURAL_UPLOAD = "Mural upload"
    NOTION_UPLOAD = "Notion upload"
    ODNOKLASSNIKI_UPLOAD = "Odnoklassniki upload"
    OK_UPLOAD = "OK upload"
    ONE_DRIVE_UPLOAD = "OneDrive upload"
    ONE_SIX_THREE_DOT_COM_UPLOAD = "163.com upload"
    ONE_TWO_SIX_DOT_COM_UPLOAD = "126.com upload"
    OPEN_TEXT_HIGHTAIL_UPLOAD = "OpenText Hightail upload"
    OTHER_DESTINATION = "Other destination"
    OUTLOOK_UPLOAD = "Outlook upload"
    OVERLEAF_UPLOAD = "Overleaf upload"
    PDF24_TOOLS_UPLOAD = "PDF24 Tools upload"
    PDF_ESCAPE_UPLOAD = "PDFescape upload"
    PDF_FILLER_UPLOAD = "pdfFiller upload"
    PDF_SIMPLI_UPLOAD = "PDFSimpli upload"
    PHOTOPEA_UPLOAD = "Photopea upload"
    PIXLR_UPLOAD = "Pixlr upload"
    PROTON_MAIL_UPLOAD = "ProtonMail upload"
    PUBLIC_LINK_FROM_CORPORATE_BOX = "Public link from corporate Box"
    PUBLIC_LINK_FROM_CORPORATE_GOOGLE_DRIVE = "Public link from corporate Google Drive"
    PUBLIC_LINK_FROM_CORPORATE_ONE_DRIVE = "Public link from corporate OneDrive"
    QQMAIL_UPLOAD = "QQMail upload"
    QZONE_UPLOAD = "Qzone upload"
    REDDIT_UPLOAD = "Reddit upload"
    REMOVABLE_MEDIA = "Removable media"
    REMOVE_DOT_BG_UPLOAD = "remove.bg upload"
    SALESFORCE_DOWNLOAD = "Download to unmonitored device from corporate Salesforce"
    SECURE_FIRM_PORTAL_UPLOAD = "Secure Firm Portal upload"
    SEJDA_UPLOAD = "Sejda upload"
    SENT_FROM_CORPORATE_GMAIL = "Sent from corporate Gmail"
    SENT_FROM_CORPORATE_OFFICE365 = "Sent from corporate Microsoft Office 365"
    SHARED_FROM_CORPORATE_BOX = "Shared from corporate Box"
    SHARED_FROM_CORPORATE_GOOGLE_DRIVE = "Shared from corporate Google Drive"
    SHARED_FROM_CORPORATE_ONE_DRIVE = "Shared from corporate OneDrive"
    SHAREFILE_UPLOAD = "Sharefile upload"
    SINA_MAIL_UPLOAD = "Sina Mail upload"
    SLACK_UPLOAD = "Slack upload"
    SMALL_PDF_UPLOAD = "SmallPDF upload"
    SMART_VAULT_UPLOAD = "SmartVault upload"
    SODA_PDF_UPLOAD = "Soda PDF upload"
    SOHU_MAIL_UPLOAD = "Sohu Mail upload"
    SOURCE_FORGE_UPLOAD = "SourceForge upload"
    STACK_OVERFLOW_UPLOAD = "Stack Overflow upload"
    STASH_UPLOAD = "Stash upload"
    SUGAR_SYNC_UPLOAD = "SugarSync upload"
    TELEGRAM_UPLOAD = "Telegram upload"
    TINY_PNG_UPLOAD = "TinyPNG upload"
    TRELLO_UPLOAD = "Trello upload"
    TUMBLR_UPLOAD = "Tumblr upload"
    TUTANOTA_UPLOAD = "Tutanota upload"
    TWITCH_UPLOAD = "Twitch upload"
    TWITTER_UPLOAD = "Twitter upload"
    UNKNOWN_DESTINATION = "Unknown destination"
    UNMONITORED_DEVICE_DOWNLOAD_BOX = (
        "Download to unmonitored device from corporate Box"
    )
    UNMONITORED_DEVICE_DOWNLOAD_GOOGLE_DRIVE = (
        "Download to unmonitored device from corporate Google Drive"
    )
    UNMONITORED_DEVICE_DOWNLOAD_ONE_DRIVE = (
        "Download to unmonitored device from corporate OneDrive"
    )
    VEED_UPLOAD = "VEED upload"
    VIMEO_UPLOAD = "Vimeo upload"
    VK_UPLOAD = "Vk upload"
    WEBEX_UPLOAD = "Webex upload"
    WE_CHAT_UPLOAD = "WeChat upload"
    WEIBO_UPLOAD = "Weibo upload"
    WE_TRANSFER_UPLOAD = "WeTransfer upload"
    WHATS_APP_UPLOAD = "WhatsApp upload"
    WIX_UPLOAD = "Wix upload"
    WORD_PRESS_UPLOAD = "WordPress upload"
    YAHOO_UPLOAD = "Yahoo upload"
    YOU_TUBE_UPLOAD = "YouTube upload"
    ZIX_UPLOAD = "Zix upload"
    ZOHO_MAIL_UPLOAD = "Zoho Mail upload"
    ZOHO_WORK_DRIVE_UPLOAD = "Zoho WorkDrive upload"
    ZOOM_UPLOAD = "Zoom upload"

    # User behavior
    FILE_MISMATCH = "File mismatch"
    OFF_HOURS = "Off hours"
    REMOTE = "Remote"
    FIRST_DESTINATION_USE = "First use of destination"
    RARE_DESTINATION_USE = "Rare use of destination"
    CONTRACT = "Contract"
    DEPARTING = "Departing"
    ELEVATED_ACCESS = "Elevated access"
    FLIGHT_RISK = "Flight risk"
    HIGH_IMPACT = "High impact"
    HIGH_RISK = "High risk"
    PERFORMANCE_CONCERNS = "Performance concerns"
    POOR_SECURITY_PRACTICES = "Poor security practices"
    SUSPICIOUS_SYSTEM_ACTIVITY = "Suspicious system activity"

    # File categories
    AUDIO = "Audio"
    DOCUMENT = "Document"
    EXECUTABLE = "Executable"
    IMAGE = "Image"
    PDF = "PDF"
    PRESENTATION = "Presentation"
    SCRIPT = "Script"
    SOURCE_CODE = "Source code"
    SPREADSHEET = "Spreadsheet"
    VIDEO = "Video"
    VIRTUAL_DISK_IMAGE = "Virtual Disk Image"
    ZIP = "Zip"


class TrustReason(_Enum):
    """Trust reasons available for filtering file events."""

    TRUSTED_DOMAIN_BROWSER_URL = "Trusted browser URL"
    TRUSTED_BROWSER_URL_PATH = "Trusted specific URL path"
    TRUSTED_DOMAIN_BROWSER_TAB_TITLE = "Trusted browser tab title"
    TRUSTED_BROWSER_TAB_INFOS = "Trusted browser URL and/or tab title"
    TRUSTED_DOMAIN_EMAIL_RECIPIENT = "Trusted email recipient"
    TRUSTED_DOMAIN_CLOUD_SYNC_USERNAME = "Trusted sync username"
    TRUSTED_SLACK_WORKSPACE = "Trusted Slack workspace"
    EVENT_PAIRING_SERVICE_MATCH = "Event matched with cloud activity"
    EVENT_PAIRING_SERVICE_ENDPOINT_MATCH = "Event matched with endpoint activity"
    DOWNLOAD_TO_A_MANAGED_DEVICE = "Download to a managed device"
    SHARED_WITH_TRUSTED_USERS = "Shared with trusted users"


class RiskSeverity(_Enum):
    """Possible Risk severity values."""

    CRITICAL = "CRITICAL"
    HIGH = "HIGH"
    MODERATE = "MODERATE"
    LOW = "LOW"
    NO_RISK_INDICATED = "NO_RISK_INDICATED"


class EventType(_Enum):
    """
    Event types available for filtering file events.
    """

    CREATED = "CREATED"
    DELETED = "DELETED"
    DOWNLOADED = "DOWNLOADED"
    EMAILED = "EMAILED"
    MODIFIED = "MODIFIED"
    OPEN = "OPEN"
    PREVIEWED = "PREVIEWED"
    PRINTED = "PRINTED"
    READ_BY_APP = "READ_BY_APP"
    RESTORE = "RESTORE"
    RULE_MATCH = "RULE_MATCH"
    UNCHANGED = "UNCHANGED"
    UNSUPPORTED = "UNSUPPORTED"


class FileType(_Enum):
    BLOCK_DEVICE = "BLOCK_DEVICE"
    BUNDLE = "BUNDLE"
    CHAR_DEVICE = "CHAR_DEVICE"
    DIR = "DIR"
    FIFO = "FIFO"
    FILE = "FILE"
    MAC_RSRC = "MAC_RSRC"
    SYMLINK = "SYMLINK"
    UNKNOWN = "UNKNOWN"
    WIN_NDS = "WIN_NDS"


class RemoteActivity(_Enum):
    FALSE = "FALSE"
    TRUE = "TRUE"
    UNKNOWN = "UNKNOWN"


class Shared(_Enum):
    FALSE = "FALSE"
    TRUE = "TRUE"



from itertools import count
from typing import Iterator
from warnings import warn

from .models import Device
from .models import DevicesPage
from .models import QueryDevicesRequest
from _incydr_sdk.enums import SortDirection
from _incydr_sdk.enums.devices import SortKeys


class DevicesClient:
    def __init__(self, parent):
        self._parent = parent
        self._v1 = None

    @property
    def v1(self):
        if self._v1 is None:
            self._v1 = DevicesV1(self._parent)
        return self._v1


class DevicesV1:
    """Client for `/v1/devices` endpoints.

    This client is deprecated. Use the Agents client instead.

    Usage example:

        >>> import incydr
        >>> client = incydr.Client(**kwargs)
        >>> client.devices.v1.get_page()
    """

    def __init__(self, parent):
        self._parent = parent

    def get_device(self, device_id: str) -> Device:
        """Get a single device.

        **Parameters**:

        * **device_id**: `str` (required) - The unique ID for the device.

        **Returns**: A [`Device`][device-model] object representing the device.

        """
        warn(
            "Devices endpoints are deprecated. Replaced by Agents.",
            DeprecationWarning,
            stacklevel=2,
        )
        response = self._parent.session.get(f"/v1/devices/{device_id}")
        return Device.parse_response(response)

    def get_page(
        self,
        active: bool = None,
        blocked: bool = None,
        page_num: int = 1,
        page_size: int = None,
        sort_dir: SortDirection = SortDirection.ASC,
        sort_key: SortKeys = SortKeys.NAME,
    ) -> DevicesPage:
        """
        Get a page of devices.

        Filter results by passing the appropriate parameters:

        **Parameters**:

        * **active**: `bool | None` - When true, return only active devices. When false, return only inactive devices. Defaults to returning both.
        * **blocked**: `bool | None` - When true, return only blocked devices. When false, return only unblocked devices. Defaults to returning both.
        * **page_num**: `int` - Page number for results, starting at 1.
        * **page_size**: `int` - Max number of results to return per page.
        * **sort_dir**: `SortDirection` - 'asc' or 'desc'. The direction in which to sort the response based on the corresponding key. Defaults to 'asc'.
        * **sort_key**: [`SortKeys`][devices-sort-keys] - One or more values on which the response will be sorted. Defaults to device name.

        **Returns**: A ['DevicesPage'][devicespage-model] object.
        """
        warn(
            "Devices endpoints are deprecated. Replaced by Agents.",
            DeprecationWarning,
            stacklevel=2,
        )
        page_size = page_size or self._parent.settings.page_size
        data = QueryDevicesRequest(
            page=page_num,
            pageSize=page_size,
            sortKey=sort_key,
            sortDirection=sort_dir,
            active=active,
            blocked=blocked,
        )
        response = self._parent.session.get("/v1/devices", params=data.dict())
        return DevicesPage.parse_response(response)

    def iter_all(
        self,
        active: bool = None,
        blocked: bool = None,
        page_size: int = None,
        sort_dir: SortDirection = SortDirection.ASC,
        sort_key: SortKeys = SortKeys.NAME,
    ) -> Iterator[Device]:
        """
        Iterate over all devices.

        Accepts the same parameters as `.get_page()` excepting `page_num`.

        **Returns**: A generator yielding individual [`Device`][device-model] objects.
        """
        warn(
            "Devices endpoints are deprecated. Replaced by Agents.",
            DeprecationWarning,
            stacklevel=2,
        )
        page_size = page_size or self._parent.settings.page_size
        for page_num in count(1):
            page = self.get_page(
                active=active,
                blocked=blocked,
                page_num=page_num,
                page_size=page_size,
                sort_dir=sort_dir,
                sort_key=sort_key,
            )
            yield from page.devices
            if len(page.devices) < page_size:
                break



from datetime import datetime
from itertools import count
from typing import Union

import requests

from ..exceptions import IncydrException
from .models import Actor
from .models import ActorFamily
from .models import ActorsPage
from .models import QueryActorsRequest
from _incydr_sdk.exceptions import DateParseError
from _incydr_sdk.queries.utils import DATE_STR_FORMAT


class ActorsV1:
    """
    Client for `/v1/actors` endpoints.

    Usage example:

        >>> import incydr
        >>> client = incydr.Client(**kwargs)
        >>> client.actors.v1.get_page()
    """

    def __init__(self, parent):
        self._parent = parent

    def get_page(
        self,
        active: bool = None,
        name_starts_with: str = None,
        name_ends_with: str = None,
        page_num: int = 1,
        page_size: int = 500,
        prefer_parent: bool = False,
    ) -> ActorsPage:
        """
        Get a page of actors.

        Filter results by passing the appropriate parameters:

        **Parameters**:

        * **active**: `bool` - Find actors with the given active state. When `True`, returns only active actors. When `False` returns only inactive actors. Defaults to `None`, returning both.
        * **name_starts_with**: `str` - Find actors whose name (e.g. username/email) starts with this text, ignoring case.
        * **name_ends_with**: `str` - Find actors whose name (e.g. username/email) ends with this text, ignoring case.
        * **page_num**: `int` - Page number for results, starting at 1.
        * **page_size**: `int` - Max number of results to return per page. Must be between 1 and 500. Defaults to 500.
        * **prefer_parent**: `bool` = Returns an actor's parent when applicable. Returns an actor themselves if they have no parent.

        **Returns**: An [`Actor`][actor-model] object representing the actor.
        """
        request_path = "/v1/actors/actor/search"

        if prefer_parent:
            request_path += "/parent"

        data = QueryActorsRequest(
            nameStartsWith=name_starts_with,
            nameEndsWith=name_ends_with,
            active=active,
            pageSize=page_size,
            page=page_num,
        )

        response = self._parent.session.get(request_path, params=data.dict())
        return ActorsPage.parse_response(response)

    def iter_all(
        self,
        active: bool = None,
        name_starts_with: str = None,
        name_ends_with: str = None,
        page_size: int = 500,
        prefer_parent: bool = False,
    ) -> ActorsPage:
        """
        Iterate over all actors.

        Accepts the same parameters as `.get_page()` excepting `page_num`.

        **Returns**: A generator yielding individual [`Actor`][actor-model] objects.
        """
        for page_num in count(1):
            page = self.get_page(
                active=active,
                name_starts_with=name_starts_with,
                name_ends_with=name_ends_with,
                page_size=page_size,
                page_num=page_num,
                prefer_parent=prefer_parent,
            )
            yield from page.actors
            if len(page.actors) < page_size:
                break

    def get_actor_by_id(self, actor_id: str, prefer_parent: bool = False) -> Actor:
        """
        Get an actor by their actor ID.

        **Parameters**:

        * **actor_id**: `str` (required) - Unique ID for the actor.
        * **prefer_parent**: `str` - Returns an actor's parent when applicable. Returns an actor themselves if they have no parent.

        **Returns**: An [`Actor`][actor-model] object representing the actor.
        """
        request_path = f"/v1/actors/actor/id/{actor_id}"
        if prefer_parent:
            request_path += "/parent"

        try:
            response = self._parent.session.get(request_path)
            return Actor.parse_response(response)
        except requests.HTTPError as err:
            if err.response.status_code == 404:
                raise ActorNotFoundError

    def get_actor_by_name(self, name: str, prefer_parent: bool = True) -> Actor:
        """
        Get an actor by their name.

        **Parameters**:

        * **name**: `str` (required) - The actor name.
        * **prefer_parent**: `str` - Returns an actor's parent when applicable. Returns an actor themselves if they have no parent. Defaults to True.

        **Returns**: An [`Actor`][actor-model] object representing the actor.
        """
        if prefer_parent:
            try:
                response = self._parent.session.get(
                    f"/v1/actors/actor/name/{name}/parent"
                )
                return Actor.parse_response(response)
            except requests.HTTPError as err:
                if err.response.status_code == 404:
                    raise ActorNotFoundError
        else:
            matches = self.get_page(name_starts_with=name)
            for actor in matches.actors:
                if actor.name == name:
                    actor_profile = self._parent.session.get(
                        f"/v1/actors/actor/id/{actor.actor_id}"
                    )
                    return Actor.parse_response(actor_profile)
            raise ActorNotFoundError(name)

    def get_family_by_member_id(self, actor_id: str) -> ActorFamily:
        """
        Get an actor family from the actor ID of any family member.

        **Parameters**:

        * **actor_id**: `str` (required) - Unique ID for an actor.

        **Returns**: An [`Actor Family`][actor-family-model] object representing the family of actors.
        """
        try:
            response = self._parent.session.get(
                f"/v1/actors/actor/id/{actor_id}/family"
            )
            return ActorFamily.parse_response(response)
        except requests.HTTPError as err:
            if err.response.status_code == 404:
                raise ActorNotFoundError

    def get_family_by_member_name(self, name: str) -> ActorFamily:
        """
        Get an actor family from the name of any family member.

        **Parameters**:

        * **name**: `str` (required) - Name of an actor.

        **Returns**: An [`ActorFamily`][actor-family-model] object representing the family of actors.
        """
        try:
            response = self._parent.session.get(f"/v1/actors/actor/name/{name}/family")
            return ActorFamily.parse_response(response)
        except requests.HTTPError as err:
            if err.response.status_code == 404:
                raise ActorNotFoundError(name)

    def update_actor(
        self,
        actor: Union[Actor, str],
        notes: str = None,
        start_date: Union[datetime, str] = None,
        end_date: Union[datetime, str] = None,
    ) -> Actor:
        """
        Update an actor.

        **Parameters**:

        Either:

        * **actor**: `str` (required) - Unique ID for an actor.
        * **notes**: `str` - Additional notes for the risk profile. Pass an empty string to clear the field.
        * **start_date**: `datetime` - The starting date for the user. Accepts a datetime object or a string in the format yyyy-MM-dd (UTC) format. Pass an empty string to clear the field.
        * **end_date**: `datetime` - The departure date for the user.  Accepts a datetime object or a string in the format yyyy-MM-dd (UTC) format.  Pass an empty string to clear the field.

        Or:

        * **actor**: An [`Actor`][actor-model] object representing the actor.

        **Returns**: An [`Actor`][actor-model] object representing the actor.
        """
        if isinstance(actor, Actor):
            return self._update_actor_from_actor(actor)
        elif notes is None and start_date is None and end_date is None:
            raise IncydrException(
                "Must provide at least one of notes or start_date or end_date arguments."
            )
        else:
            return self._update_actor_from_args(
                actor_id=actor, start_date=start_date, end_date=end_date, notes=notes
            )

    def _update_actor_from_actor(self, actor: Actor) -> Actor:
        try:
            response = self._parent.session.patch(
                f"/v1/actors/actor/id/{actor.actor_id}",
                json={
                    "notes": actor.notes,
                    "startDate": actor.start_date,
                    "endDate": actor.end_date,
                },
            )
            return Actor.parse_response(response)
        except requests.HTTPError as err:
            if err.response.status_code == 404:
                raise ActorNotFoundError(actor.actor_id)

    def _update_actor_from_args(
        self,
        actor_id: str,
        notes: str = None,
        start_date: Union[datetime, str] = None,
        end_date: Union[datetime, str] = None,
    ) -> Actor:
        request_body = {}
        if notes is not None:
            request_body["notes"] = None if notes == "" else notes
        if start_date is not None:
            request_body["startDate"] = (
                None if start_date == "" else _create_date(start_date)
            )
        if end_date is not None:
            request_body["endDate"] = None if end_date == "" else _create_date(end_date)
        try:
            response = self._parent.session.patch(
                f"/v1/actors/actor/id/{actor_id}", json=request_body
            )
            return Actor.parse_response(response)
        except requests.HTTPError as err:
            if err.response.status_code == 404:
                raise ActorNotFoundError(actor_id)


class ActorsClient:
    def __init__(self, parent):
        self._parent = parent
        self._v1 = None

    @property
    def v1(self):
        if self._v1 is None:
            self._v1 = ActorsV1(self._parent)
        return self._v1


class ActorNotFoundError(IncydrException):
    """Raised when a watchlist with the matching type or title is not found."""

    def __init__(self, identifier=None):
        self.identifier = identifier
        self.message = (
            (f"Actor Not Found Error: No results for actors matching '{identifier}'.")
            if identifier
            else "Actor Not Found Error: No results for specified actor(s)."
        )
        super().__init__(self.message)


def _create_date(date: Union[datetime, str]):
    if not isinstance(date, datetime):
        try:
            date = datetime.strptime(date, DATE_STR_FORMAT)
        except ValueError:
            raise DateParseError(
                date,
                msg=f"DateParseError: Error parsing time data. Date '{date}' does not match format {DATE_STR_FORMAT}.",
            )
    return date.strftime(DATE_STR_FORMAT)



from typing import List
from typing import Optional

from pydantic import BaseModel
from pydantic import Field

from _incydr_sdk.core.models import ResponseModel


class DepartmentsPage(ResponseModel):
    """
    A model representing a list of departments.

    **Fields**:

    * **departments**: `List[str]` - The list of `n` department names retrieved by the query, where `n=page_size`.
    * **total_count**: `int` - Total count of departments retrieved by the query.
    """

    departments: Optional[List[str]] = None
    total_count: Optional[int] = Field(
        None,
        description="The total count of all departments.",
        example=10,
        alias="totalCount",
    )


class GetPageRequest(BaseModel):
    page: int = 1
    page_size: int = None
    name: str = None



from typing import Optional

import click
from rich.table import Table

from _incydr_cli import console
from _incydr_cli import logging_options
from _incydr_cli import render
from _incydr_cli.cmds.options.output_options import columns_option
from _incydr_cli.cmds.options.output_options import single_format_option
from _incydr_cli.cmds.options.output_options import SingleFormat
from _incydr_cli.cmds.options.output_options import table_format_option
from _incydr_cli.cmds.options.output_options import TableFormat
from _incydr_cli.core import IncydrCommand
from _incydr_cli.core import IncydrGroup
from _incydr_cli.render import measure_renderable
from _incydr_sdk.alert_rules.client import MissingUsernameCriterionError
from _incydr_sdk.alert_rules.models.response import RuleDetails
from _incydr_sdk.core.client import Client
from _incydr_sdk.utils import list_as_panel
from _incydr_sdk.utils import model_as_card


MAX_USER_DISPLAY_COUNT = 25


@click.group(cls=IncydrGroup)
@logging_options
def alert_rules():
    """View and manage alert rules."""


@alert_rules.command("list", cls=IncydrCommand)
@table_format_option
@columns_option
@logging_options
def list_(
    format_: TableFormat,
    columns: Optional[str],
):
    """
    List all rules.
    """
    client = Client()
    rules = client.alert_rules.v2.iter_all()

    if format_ == TableFormat.table:
        if not columns:
            columns = [
                "id",
                "name",
                "severity",
                "is_enabled",
                "created_at",
                "created_by",
                "modified_at",
                "modified_by",
                "description",
            ]
        render.table(RuleDetails, rules, columns=columns, flat=False)
    elif format_ == TableFormat.csv:
        render.csv(RuleDetails, rules, columns=columns, flat=True)
    elif format_ == TableFormat.json_pretty:
        for rule in rules:
            console.print_json(rule.json())
    else:
        for rule in rules:
            click.echo(rule.json())


@alert_rules.command(cls=IncydrCommand)
@single_format_option
@click.argument("rule-id")
@logging_options
def show(
    rule_id: str,
    format_: SingleFormat,
):
    """
    Show details for a single rule.

    If using `rich`, also retrieve the username filter for the rule (if it exists).
    """
    client = Client()
    rule = client.alert_rules.v2.get_rule(rule_id)

    if format_ == SingleFormat.json_pretty:
        console.print_json(rule.json())
        return
    if format_ == SingleFormat.json_lines:
        click.echo(rule.json())
        return

    t = Table(title=f"Alert Rule {rule.id}")
    t.add_column("Info")
    tables = []

    try:
        username_filter = client.alert_rules.v2.get_users(rule_id)
        # Pretty sure there is only ever one "user" object whose aliases indicate the usernames on the filter
        usernames = []
        for user in username_filter.users:
            usernames.extend(user.aliases)
        if usernames:
            tables.append(list_as_panel(usernames[:MAX_USER_DISPLAY_COUNT]))
            t.add_column(
                f"{'Include' if username_filter.mode == '0' else 'Exclude'}d Usernames"
            )
    except MissingUsernameCriterionError:
        pass

    if rule.notifications:
        tables.append(model_as_card(rule.notifications))
        t.add_column("Notifications")
    if rule.education:
        tables.append(model_as_card(rule.education))
        t.add_column("Education")
    if rule.vectors:
        tables.append(model_as_card(rule.vectors))
        t.add_column("Vectors")
    if rule.filters:
        tables.append(model_as_card(rule.filters))
        t.add_column("Filters")

    t.add_row(
        model_as_card(
            rule,
            include=[
                "name",
                "description",
                "severity",
                "is_enabled",
                "created_at",
                "created_by",
                "modified_at",
                "modified_by",
                "is_system_rule",
            ],
        ),
        *tables,
    )

    with console.pager():
        # expand console and table so no values get truncated due to size of console, since we're using a pager
        console.width = t.width = measure_renderable(t)
        console.print(t, crop=False, soft_wrap=False, overflow="fold")


@alert_rules.command(cls=IncydrCommand)
@click.argument("rule-ids")
@logging_options
def enable(
    rule_ids: str,
):
    """
    Enable a single rule or a set of rules.

    Where RULE-IDS is a comma-delimited list of rule IDs to enable.
    """
    client = Client()
    client.alert_rules.v2.enable_rules([r.strip() for r in rule_ids.split(",")])
    console.print("Successfully enabled rule(s).")


@alert_rules.command(cls=IncydrCommand)
@click.argument("rule-ids")
@logging_options
def disable(
    rule_ids: str,
):
    """
    Disable a single rule or a set of rules.
    """
    client = Client()
    client.alert_rules.v2.disable_rules([r.strip() for r in rule_ids.split(",")])
    console.print("Successfully disabled rule(s).")


@alert_rules.command(cls=IncydrCommand)
@click.argument("rule-id")
@logging_options
def remove_all_users(
    rule_id: str,
):
    """
    Remove ALL users from a rule's username filter.

    Note that the removed users could become either included or excluded from the rule,
    depending on the rule's configuration.
    """
    client = Client()
    client.alert_rules.v2.remove_all_users(rule_id)
    console.print(f"Successfully removed all users from rule '{rule_id}'.")


@alert_rules.command(cls=IncydrCommand)
@click.argument("rule-id")
@single_format_option
@logging_options
def list_users(
    rule_id: str,
    format_: SingleFormat,
):
    """
    Lists the usernames on the rule's username filter.

    Note that users could either be included on or excluded from the rule depending on the rule's configuration.
    """
    client = Client()
    try:
        username_filter = client.alert_rules.v2.get_users(rule_id)
    except MissingUsernameCriterionError:
        console.print(f"No results found for rule {rule_id}")
        return

    usernames = []
    for user in username_filter.users:
        usernames.extend(user.aliases)

    if not usernames:
        console.print(f"No results found for rule {rule_id}")
        return

    if format_ == SingleFormat.rich:
        console.print(
            list_as_panel(
                usernames,
                expand=False,
                title=f"{'Include' if username_filter.mode == '0' else 'Exclude'}d Usernames",
            )
        )
    elif format_ == SingleFormat.json_pretty:
        console.print_json(username_filter.json())
    else:
        click.echo(username_filter.json())



from datetime import datetime
from typing import List
from typing import Optional

from pydantic import BaseModel
from pydantic import Field

from _incydr_sdk.core.models import ResponseModel
from _incydr_sdk.enums import SortDirection
from _incydr_sdk.enums.devices import SortKeys


class Device(ResponseModel):
    """
    A model representing a device.

    **Fields**:

    * **device_id**: `str` - The globally unique ID (guid) for this device.
    * **legacy_device_id**: `str` - The device ID to use for older console-based APIs that require a device ID.
    * **name**: `str` - The device name.
    * **os_hostname**: `str` - The device hostname according to the device's OS.
    * **status**: `str` - The device status. One of [Active, Deactivated, Blocked, Deauthorized (Active/Deactivated followed by optional Blocked and/or Deauthorized).
    * **active**: `bool` - Whether or not the device is active. If true, the device will show up on reports, etc.
    * **blocked**: `bool` - Whether or not the device is blocked.  If true, restores and logins are disabled.
    * **alert_state**: `int` - The device's alert state. One of [0=ok, 1=connection warning, 2=connection critical].
    * **user_id**: `str` - The globally unique ID for this user.
    * **legacy_user_id**: `str` - The user ID to use for older console-based APIs that require a user ID.
    * **org_id**: `str` - An ID for the Code42 organization of the user owning this device.
    * **legacy_org_id**: `str` - The org ID to use for older console-based APIs that require an org ID.
    * **org_guid**: `str` - The globally unique org ID.  This is the org identifier that should be used for all org-related API actions.
    * **external_reference**: `str` - Optional external reference information, such as a serial number, asset tag, employee ID, or help desk issue ID.
    * **notes**: `str` - Optional descriptive information for the device.
    * **last_connected**: `datetime` - The last day and time this device was connected to the server.
    * **os_name**: `str` - The device's operating system name. Ex: Windows*, Mac OS X, Linux, Android, iOS, SunOS, etc.
    * **os_version**: `str` - The device's operating system version. Ex: 10.5.1, 6.2, etc.
    * **address: `str` - The device's internal IP address and port. Ex: 192.168.42.1:4282
    * **remote_address**: `str` - The device's external IP address and port. Example: 171.22.110.41:13958
    * **time_zone**: `str` - The device's time zone. Ex: Asia/Calcutta
    * **version**: `str` - The device product display version.
    * **build**: `int` - The device build version long number, will only be applicable to CP4/SP devices.
    * **creation_date**: `datetime` - The date and time this device was created.
    * **modification_date**: `datetime` - The date and time this device was last modified.
    * **login_date**: `datetime` - The date and time this device was last logged in.
    """

    device_id: Optional[str] = Field(
        alias="deviceId",
        description="A globally unique ID (guid) for this device.",
    )
    legacy_device_id: Optional[str] = Field(
        alias="legacyDeviceId",
        description="The device ID to use for older console-based APIs that require a device Id.",
    )
    name: Optional[str] = Field(None, description="Device name.")
    os_hostname: Optional[str] = Field(
        alias="osHostname",
        description="Device Hostname according to the device's OS.",
    )
    status: Optional[str] = Field(
        description="Device status. Values: Active, Deactivated, Blocked, Deauthorized (Active/Deactivated followed by optional Blocked and/or Deauthorized)",
    )
    active: Optional[bool] = Field(
        description="True means the device will show up on reports, etc.",
    )
    blocked: Optional[bool] = Field(
        description="True means device continues backing up, but restores and logins are disabled.",
    )
    alert_state: Optional[int] = Field(
        alias="alertState",
        description="0=ok, 1=connection warning, 2=connection critical",
    )
    user_id: Optional[str] = Field(
        alias="userId",
        description="A globally unique ID for this user.",
    )
    legacy_user_id: Optional[str] = Field(
        alias="legacyUserId",
        description='The user ID to use for older console-based APIs that require a user Id.\r\nIf your endpoint domain starts with "console" instead of "api", use this Id for endpoints that require a userId.',
    )
    org_id: Optional[str] = Field(
        alias="orgId",
        description="An ID for the Code42 organization of the user owning this device.",
    )
    legacy_org_id: Optional[str] = Field(
        alias="legacyOrgId",
        description='The org ID to use for older console-based APIs that require an org Id.\r\nIf your endpoint domain starts with "console" instead of "api", use this Id for endpoints that require an orgId.',
    )
    org_guid: Optional[str] = Field(
        alias="orgGuid", description="The globally unique org ID."
    )
    external_reference: Optional[str] = Field(
        alias="externalReferenceInfo",
        description="Optional external reference information, such as a serial number, asset tag, employee ID, or help desk issue ID.",
    )
    notes: Optional[str] = Field(None, description="Optional descriptive information.")
    last_connected: Optional[datetime] = Field(
        alias="lastConnected",
        description="The last day and time this device was connected to the server.",
    )
    os_name: Optional[str] = Field(
        alias="osHostname",
        description="Operating system name. Values: Windows*, Mac OS X, Linux, Android, iOS, SunOS, etc",
    )
    os_version: Optional[str] = Field(
        alias="osVersion",
        description="Operating system version. Values: 10.5.1, 6.2, etc",
    )
    os_arch: Optional[str] = Field(
        alias="osArch",
        description="Hardware architecture. Values: i386, amd64, sparc, x86, etc",
    )
    address: Optional[str] = Field(
        description="Internal IP address and port. Example: 192.168.42.1:4282",
    )
    remote_address: Optional[str] = Field(
        alias="remoteAddress",
        description="External IP address and port. Example: 171.22.110.41:13958",
    )
    time_zone: Optional[str] = Field(
        alias="timeZone",
        description="Examples: Australia/Canberra, Asia/Calcutta",
    )
    version: Optional[str] = Field(None, description="Device product display version.")
    build: Optional[int] = Field(
        description="Device build version long number, will only be applicable to CP4/SP devices.",
    )
    creation_date: Optional[datetime] = Field(
        alias="creationDate",
        description="Date and time this device was created.",
    )
    modification_date: Optional[datetime] = Field(
        alias="modificationDate",
        description="Date and time this device was last modified.",
    )
    login_date: Optional[datetime] = Field(
        alias="loginDate",
        description="Date and time this device was last logged in.",
    )


class DevicesPage(ResponseModel):
    """
    A model representing a page of `Device` objects.

    **Fields**:

    * **devices**: `List[Device]` - The list of `n` number of devices retrieved from the query, where `n=page_size`.
    * **total_count**: `int` - Total count of devices found by query.
    """

    total_count: Optional[int] = Field(
        alias="totalCount", description="The total number of devices"
    )
    devices: Optional[List[Device]] = Field(description="A list of devices")


class QueryDevicesRequest(BaseModel):
    active: Optional[bool]
    blocked: Optional[bool]
    page: Optional[int] = 1
    pageSize: Optional[int] = 100
    sortDirection: SortDirection = SortDirection.ASC
    sortKey: SortKeys = SortKeys.NAME

    class Config:
        use_enum_values = True



from datetime import date
from datetime import datetime
from datetime import timedelta
from typing import List
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import conint
from pydantic import Field
from pydantic import root_validator
from pydantic import StrictBool

from _incydr_sdk.core.models import Model
from _incydr_sdk.enums.alerts import AlertSeverity
from _incydr_sdk.enums.alerts import AlertState
from _incydr_sdk.enums.alerts import AlertTerm
from _incydr_sdk.enums.alerts import Operator
from _incydr_sdk.enums.alerts import RiskSeverity
from _incydr_sdk.queries.utils import parse_ts_to_ms_str


_term_enum_map = {
    AlertTerm.STATE: AlertState,
    AlertTerm.RISK_SEVERITY: RiskSeverity,
    AlertTerm.SEVERITY: AlertSeverity,
}


class Filter(BaseModel):
    term: AlertTerm
    operator: Operator
    value: Optional[Union[StrictBool, int, str]]

    class Config:
        use_enum_values = True

    @root_validator(pre=True)
    def _validate_enums(cls, values: dict):  # noqa `root_validator` is a classmethod
        term = values.get("term")
        operator = values.get("operator")
        value = values.get("value")

        # make sure `term` is valid enum value
        AlertTerm(term)
        if term == "HasAuthSignificantWatchlist" and not isinstance(value, bool):
            raise ValueError("HasAuthSignificantWatchlist requires a boolean value.")

        enum = _term_enum_map.get(term)
        if enum:
            enum(value)

        if operator in (Operator.IS, Operator.IS_NOT):
            if not isinstance(value, (str, int)):
                raise ValueError(
                    f"`IS` and `IS_NOT` filters require a `str | int` value, got term={term}, operator={operator}, value={value}."
                )

        return values


class FilterGroup(BaseModel):
    filterClause: str = "AND"
    filters: Optional[List[Filter]]


class AlertQuery(Model):
    """
    Class to build an alert query. Use the class methods to attach additional filter operators.

    Usage examples:

    Construct a query that finds all alerts created in the past 10 days that are now marked `RESOLVED`:

        >>> import incydr
        >>> from datetime import timedelta
        >>> from incydr.enums.alerts import AlertState, AlertTerm
        >>> query = incydr.AlertQuery(start_date=timedelta(days=10)).equals(AlertTerm.STATE, AlertState.RESOLVED)

    Construct a query that finds alerts triggered by actor `user@example.com` with a risk severity of either `HIGH` or
    `CRITICAL`:

        >>> import incydr
        >>> query = incydr.AlertQuery().equals("Actor", "user@example.com").equals("RiskSeverity", ["HIGH", "CRITICAL"])

    **Parameters**:

     * **start_date**: `int | float | str | datetime | timedelta` - Start of the date range to query for alerts.
        Accepts `int|float` unix timestamp, string representation of the date (`%Y-%m-%d %H:%M:%S` or `%Y-%m-%d`
        formats), or `datetime` object for absolute dates, or if a `timedelta` is provided it will be relative to the
        current time when the query object was instantiated. Defaults to None.
     * **end_date**: `int | float | str | datetime` - End of the date range to query for alerts. Defaults to
        None.
     * **on**: `str | date | datetime` - Retrieve alerts that occurred on a specific date (incompatible with either
        `start_date` or `end_date` arguments).
    """

    tenant_id: str = Field(None, alias="tenantId")
    group_clause: str = Field("AND", alias="groupClause")
    groups: Optional[List[FilterGroup]]
    page_num: int = Field(0, alias="pgNum")
    page_size: conint(gt=0, le=500) = Field(100, alias="pgSize")
    sort_dir: str = Field("DESC", alias="srtDirection")
    sort_key: AlertTerm = Field("CreatedAt", alias="srtKey")

    def __init__(
        self,
        start_date: Union[datetime, timedelta, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        on: Union[date, datetime, int, float, str] = None,
        **kwargs,
    ):
        groups = kwargs.get("groups") or []
        if on and any((start_date, end_date)):
            raise ValueError(
                "cannot use 'on' argument with 'start_date' or 'end_date' arguments."
            )
        if start_date or end_date or on:
            date_filter = _create_date_range_filter_group(start_date, end_date, on)
            groups.append(date_filter)
        kwargs["groups"] = groups
        super().__init__(**kwargs)

    def equals(self, term: str, values: Union[str, List[str]]):
        """
        Adds an `equals` filter to the query. The opposite of the `not_equals` filter.

        Causes the query to return alerts where the field corresponding to the term equals the indicated value(s).

        Example: `AlertQuery().equals('Status', 'PENDING')` will return all alerts with a status of 'PENDING'.

        **Parameters**:

        * **term**: `str` - The term which corresponds to an alert field. List of valid terms can be found in the
            [`incydr.enums.alerts.AlertTerm`][alert-terms] enum object.
        * **values**: `str`, `List[str]` - The value(s) for the term to match.
        """
        if isinstance(values, (str, bool)):
            values = [values]
        if len(values) < 1:
            raise ValueError("equals() requires at least one value.")
        filters = [Filter(term=term, operator=Operator.IS, value=val) for val in values]
        filter_group = FilterGroup(
            filters=filters,
            filterClause="OR" if len(values) > 1 else "AND",
        )
        self.groups.append(filter_group)
        return self

    def not_equals(self, term: str, values: Union[str, List[str]]):
        """
        Adds an `not_equals` filter to the query. The opposite of the `equals` filter.

        When passed as part of a query, returns alerts where the field corresponding to the filter term does not equal
        the indicated value(s).

        Example:
            `AlertQuery().not_equals('State', 'RESOLVED')` creates a query which will return alerts that are not in a
            `RESOLVED` state.

        **Parameters**:

        * **term**: `str` - The term which corresponds to an alert field.
        * **values**: `str`, `List[str]` - The value(s) for the term to not match.
        """

        if isinstance(values, str):
            values = [values]
        if len(values) < 1:
            raise ValueError("not_equals() requires at least one value.")
        filters = [
            Filter(term=term, operator=Operator.IS_NOT, value=val) for val in values
        ]
        filter_group = FilterGroup(
            filters=filters,
            filterClause="OR" if len(values) > 1 else "AND",
        )
        self.groups.append(filter_group)
        return self

    def contains(self, term: str, value: str):
        """
        Adds a `contains` filter to the query. The opposite of the `does_not_contain` filter.

        When passed as part of a query, returns alerts where the field corresponding to the filter term contains the
        provided string value.

        Example:
            `AlertQuery().contains('Description', 'removable media')` creates a query which will return alerts where the
            alert rule description contains the string "removable media".

        **Parameters**:

        * **term**: `str` - The term which corresponds to an alert field.
        * **value**: `str` - The value for the term to match.
        """
        self.groups.append(
            FilterGroup(
                filters=[Filter(term=term, operator=Operator.CONTAINS, value=value)]
            )
        )
        return self

    def does_not_contain(self, term: str, value: str):
        """
        Adds a `does_not_contain` filter to the query. The opposite of the `contains` filter.

        When passed as part of a query, returns alerts where the field corresponding to the filter term does not contain
        the provided string value.

        Example:
            `AlertQuery().does_not_contain('Description', 'removable media')` creates a query which will return alerts
            where the alert rule description does not contain the string "removable media".

        **Parameters**:

        * **term**: `str` - The term which corresponds to an alert field.
        * **value**: `str` - The value for the term to not match.
        """
        self.groups.append(
            FilterGroup(
                filters=[
                    Filter(term=term, operator=Operator.DOES_NOT_CONTAIN, value=value)
                ]
            )
        )
        return self

    def matches_any(self):
        """
        Sets operator to combine multiple filters with `OR`.

        Returns alerts that match at least one of the filters in the query.

        Default operator is `AND`, which returns only alerts that match _all_ filters in the query.
        """
        self.group_clause = "OR"
        return self


def _create_date_range_filter_group(start_date, end_date, on):

    filters = []

    if isinstance(start_date, timedelta):
        start_date = datetime.utcnow() - start_date
    if on:
        filters.append(
            Filter(
                term=AlertTerm.CREATED_AT,
                operator=Operator.ON,
                value=parse_ts_to_ms_str(on),
            )
        )
        return FilterGroup(filters=filters)
    if start_date:
        filters.append(
            Filter(
                term=AlertTerm.CREATED_AT,
                operator=Operator.ON_OR_AFTER,
                value=parse_ts_to_ms_str(start_date),
            )
        )

    if end_date:
        filters.append(
            Filter(
                term=AlertTerm.CREATED_AT,
                operator=Operator.ON_OR_BEFORE,
                value=parse_ts_to_ms_str(end_date),
            )
        )
    return FilterGroup(filters=filters)



from typing import Optional

import click
from rich.panel import Panel

from _incydr_cli import console
from _incydr_cli import logging_options
from _incydr_cli import render
from _incydr_cli.cmds.options.output_options import columns_option
from _incydr_cli.cmds.options.output_options import single_format_option
from _incydr_cli.cmds.options.output_options import SingleFormat
from _incydr_cli.cmds.options.output_options import table_format_option
from _incydr_cli.cmds.options.output_options import TableFormat
from _incydr_cli.cmds.options.profile_filter_options import profile_filter_options
from _incydr_cli.cmds.options.utils import user_lookup_callback
from _incydr_cli.cmds.utils import deprecation_warning
from _incydr_cli.core import incompatible_with
from _incydr_cli.core import IncydrCommand
from _incydr_cli.core import IncydrGroup
from _incydr_sdk.core.client import Client
from _incydr_sdk.exceptions import DateParseError
from _incydr_sdk.risk_profiles.models import RiskProfile
from _incydr_sdk.utils import model_as_card

# Deprecated September 2024.
DEPRECATION_TEXT = "DeprecationWarning: Risk Profile commands are deprecated. Use the 'incydr actors' command group instead."


@click.group(cls=IncydrGroup)
@logging_options
def risk_profiles():
    """DEPRECATED. Use the Actors command group instead. View and manage risk profiles."""
    deprecation_warning(DEPRECATION_TEXT)


@risk_profiles.command("list", cls=IncydrCommand)
@table_format_option
@columns_option
@profile_filter_options
@logging_options
def list_(
    format_: TableFormat,
    columns: Optional[str],
    manager: Optional[str],
    title: Optional[str],
    division: Optional[str],
    department: Optional[str],
    employment_type: Optional[str],
    country: Optional[str],
    region: Optional[str],
    locality: Optional[str],
    active: Optional[bool],
    deleted: Optional[bool],
    support_user: Optional[bool],
):
    """
    List risk profiles.
    """
    client = Client()
    profiles = client.risk_profiles.v1.iter_all(
        manager_id=manager,
        title=title,
        division=division,
        department=department,
        employment_type=employment_type,
        country=country,
        region=region,
        locality=locality,
        active=active,
        deleted=deleted,
        support_user=support_user,
    )

    if format_ == TableFormat.csv:
        render.csv(RiskProfile, profiles, columns=columns, flat=True)
    elif format_ == TableFormat.table:
        columns = columns or [
            "username",
            "user_id",
            "display_name",
            "cloud_aliases",
            "active",
            "start_date",
            "end_date",
        ]
        render.table(RiskProfile, profiles, columns=columns, flat=False)
    else:
        printed = False
        if format_ == TableFormat.json_pretty:
            for p in profiles:
                printed = True
                console.print_json(p.json())
        else:
            for p in profiles:
                printed = True
                click.echo(p.json())
        if not printed:
            console.print("No results found.")


@risk_profiles.command(cls=IncydrCommand)
@click.argument("user", callback=user_lookup_callback)
@single_format_option
@logging_options
def show(
    user: str,
    format_: SingleFormat,
):
    """
    Show details for a risk profile.

    Accepts a user ID or a username.  Performs an additional lookup if a username is passed.
    """
    client = Client()
    profile = client.risk_profiles.v1.get_risk_profile(user)
    if format_ == SingleFormat.rich and client.settings.use_rich:
        console.print(
            Panel.fit(model_as_card(profile), title=f"Risk Profile {profile.username}")
        )
    elif format_ == SingleFormat.json_pretty:
        console.print_json(profile.json())
    else:
        click.echo(profile.json())


@risk_profiles.command(cls=IncydrCommand)
@click.argument("user", callback=user_lookup_callback)
@click.option(
    "--start-date",
    default=None,
    help="Update a user's starting date. Accepts a date in yyyy-MM-dd (UTC) format.",
)
@click.option(
    "--end-date",
    default=None,
    help="Update a user's departure date. Accepts a date in yyyy-MM-dd (UTC) format.",
)
@click.option(
    "--notes", default=None, help="Update the optional notes on a user's profile."
)
@click.option(
    "--clear-start-date",
    is_flag=True,
    default=False,
    help="Clear the start date on a user's profile. Incompatible with --start-date.",
    cls=incompatible_with("start_date"),
)
@click.option(
    "--clear-end-date",
    is_flag=True,
    default=False,
    help="Clear the end date on a user's profile. Incompatible with --end-date.",
    cls=incompatible_with("end_date"),
)
@click.option(
    "--clear-notes",
    is_flag=True,
    default=False,
    help="Clear the notes on a user's profile. Incompatible with --notes.",
    cls=incompatible_with("notes"),
)
@logging_options
def update(
    user,
    start_date=None,
    end_date=None,
    notes=None,
    clear_start_date=None,
    clear_end_date=None,
    clear_notes=None,
):
    """
    Update a risk profile.

    Accepts a user ID or a username.  Performs an additional lookup if a username is passed.
    """
    if not any(
        [start_date, end_date, notes, clear_start_date, clear_end_date, clear_notes]
    ):
        raise click.UsageError(
            "At least one of --start-date, --end-date, or --notes, or one of their corresponding clear flags, "
            "is required to update a risk profile."
        )

    client = Client()

    if clear_start_date:
        start_date = ""
    if clear_end_date:
        end_date = ""
    if clear_notes:
        notes = ""

    try:
        updated_profile = client.risk_profiles.v1.update(
            user, notes, start_date, end_date
        )
        if client.settings.use_rich:
            console.print(
                Panel.fit(model_as_card(updated_profile), title="Updated Risk Profile")
            )
        else:
            console.print(updated_profile.json(), highlight=False)
    except DateParseError as err:
        raise err



from typing import List
from typing import Optional

from pydantic import Field

from _incydr_sdk.core.models import ResponseModel
from _incydr_sdk.file_events.models.response import FileEventsPage
from _incydr_sdk.sessions.models.models import Alert
from _incydr_sdk.sessions.models.models import ContentInspectionResult
from _incydr_sdk.sessions.models.models import Note
from _incydr_sdk.sessions.models.models import RiskIndicator
from _incydr_sdk.sessions.models.models import Score
from _incydr_sdk.sessions.models.models import State


class Session(ResponseModel):
    """
    A model representing a session summary.

    **Fields**:

    * **actor_id**: `str` The ID of the actor that generated the session.
    * **begin_time**: `datetime` The date and time when this session began.
    * **content_inspection_results**: `List[ContentInspectionResult]` The results of content inspection.
    * **context_summary**: `str` An English summary of the contextual aspects of this session is any were identified.
    * **critical_events**: `int` The number of events in the session with a critical risk severity.
    * **end_time**: `datetime` The date and time when this session ended.
    * **exfiltration_summary**: `str` An English summary of the exfiltration (if any) in this session.
    * **first_observed**: `datetime` The date and time at which this session was first observed.
    * **high_events**: `int` The number of events in the session with a high risk severity.
    * **last_updated**: `datetime` The date and time at which this session was last updated.
    * **low_events**: `int` The number of events in the session with a low risk severity
    * **moderate_events**: `int` The number of events in the session with a moderate risk severity.
    * **no_risk_events**: `int` The number of events in the session with a no risk severity.
    * **notes**: `str` The set of notes associated with this session.
    * **risk_indicators**: `List[str]` The list of risk indicator/weight combinations observed in this session.
    * **scores**: `str` The history of all score assignments for this session.
    * **session_id**: `str` The session ID.
    * **states**: `str` The history of states this session has been in.
    * **tenant_id**: `str` The tenant ID.
    * **triggered_alerts**: `str` The list of all alerts that were triggered by activity in this session.
    """

    actor_id: Optional[str] = Field(alias="actorId")
    begin_time: Optional[int] = Field(alias="beginTime")
    content_inspection_results: Optional[ContentInspectionResult] = Field(
        alias="contentInspectionResults"
    )
    context_summary: Optional[str] = Field(alias="contextSummary")
    critical_events: Optional[int] = Field(alias="criticalEvents")
    end_time: Optional[int] = Field(alias="endTime")
    exfiltration_summary: Optional[str] = Field(alias="exfiltrationSummary")
    first_observed: Optional[int] = Field(alias="firstObserved")
    high_events: Optional[int] = Field(alias="highEvents")
    last_updated: Optional[int] = Field(alias="lastUpdated")
    low_events: Optional[int] = Field(alias="lowEvents")
    moderate_events: Optional[int] = Field(alias="moderateEvents")
    no_risk_events: Optional[int] = Field(alias="noRiskEvents")
    notes: Optional[List[Note]]
    risk_indicators: Optional[List[RiskIndicator]] = Field(alias="riskIndicators")
    scores: Optional[List[Score]]
    session_id: Optional[str] = Field(alias="sessionId")
    states: Optional[List[State]]
    tenant_id: Optional[str] = Field(alias="tenantId")
    triggered_alerts: Optional[List[Alert]] = Field(alias="triggeredAlerts")
    user_id: Optional[str] = Field(None, alias="userId")


class SessionsPage(ResponseModel):
    """
    A model representing a page of sessions.

    **Fields**:

    * **items**: `List[Session]` The list of sessions returned by the query.
    * **total_count**: `int` The total count of sessions returned by the query.
    """

    items: List[Session] = Field(None)
    total_count: int = Field(None, alias="totalCount")


class SessionEvents(ResponseModel):
    """
    The wrapped file event search response returned when retrieving the events attached to a session.
    """

    query_result: FileEventsPage = Field(alias="queryResult")



import json
import os
from contextlib import nullcontext
from datetime import timezone
from hashlib import md5
from itertools import count
from typing import Optional

import click
from pydantic import Field

from _incydr_cli import console
from _incydr_cli import get_user_project_path
from _incydr_cli import logging_options
from _incydr_cli import render
from _incydr_cli.cmds.options.audit_log_filter_options import filter_options
from _incydr_cli.cmds.options.output_options import columns_option
from _incydr_cli.cmds.options.output_options import output_options
from _incydr_cli.cmds.options.output_options import TableFormat
from _incydr_cli.cmds.options.utils import checkpoint_option
from _incydr_cli.cmds.utils import warn_interrupt
from _incydr_cli.core import IncydrCommand
from _incydr_cli.core import IncydrGroup
from _incydr_cli.cursor import CursorStore
from _incydr_cli.logger import get_server_logger
from _incydr_sdk.audit_log.models import DateRange
from _incydr_sdk.audit_log.models import QueryAuditLogRequest
from _incydr_sdk.core.client import Client
from _incydr_sdk.core.models import Model
from _incydr_sdk.queries.utils import parse_str_to_dt
from _incydr_sdk.queries.utils import parse_ts_to_posix_ts


@click.group(cls=IncydrGroup)
@logging_options
def audit_log():
    """View audit log events."""


@audit_log.command("search", cls=IncydrCommand)
@checkpoint_option
@click.option(
    "--format",
    "-f",
    "format_",
    type=click.Choice(
        [TableFormat.csv, TableFormat.json_lines, TableFormat.json_pretty]
    ),
    help="Format to print result. One of 'json-pretty', 'json-lines', or 'csv. "
    "'table' format is unavailable due to long processing times for very large data sets."
    "If environment has INCYDR_USE_RICH=false set, defaults to 'json-lines', else defaults to 'json-pretty'."
    "CSV output includes limited fields, use audit-log download for a more comprehensive CSV download.",
    default=TableFormat.json_pretty,
)
@filter_options
@columns_option
@output_options
@logging_options
def search(
    format_: TableFormat,
    columns: Optional[str],
    output: Optional[str],
    certs: Optional[str],
    ignore_cert_validation: Optional[bool],
    start: Optional[str],
    end: Optional[str],
    actor_ids: Optional[str],
    actor_ip_addresses: Optional[str],
    actor_names: Optional[str],
    event_types: Optional[str],
    resource_ids: Optional[str],
    user_types: Optional[str],
    checkpoint_name: Optional[str],
):
    """
    Search audit log events.  Returns all events that match the search criteria with paging.

    Defaults to searching for most recent events.

    Results will be output to the console by default, use the `--output` option to send data to a server.

    Checkpointing is available through the `--checkpoint <checkpoint-name>` option and will only return new results
    on subsequent queries with that same checkpoint. Checkpointing filters by timestamp, additional filter
    options will need to be included in each run.
    """
    client = Client()
    cursor = _get_cursor_store(client.settings.api_client_id)

    if output:
        format_ = TableFormat.json_lines

    # skip pydantic modeling when output will just be json
    if format_ in (TableFormat.json_pretty, TableFormat.json_lines):

        def yield_all_events(request_):
            for page_num in count(0):
                request_.page = page_num
                page = client.session.post(
                    "/v1/audit/search-audit-log", json=request_.dict()
                )
                events = page.json().get("events")
                yield from events
                if len(events) < request_.pageSize:
                    break

    else:

        def yield_all_events(request_):
            for page_num in count(0):
                request_.page = page_num
                page = client.session.post(
                    "/v1/audit/search-audit-log", json=request_.dict()
                )
                events = page.json().get("events")
                for e in events:
                    yield DefaultAuditEvent.parse_obj(e)
                if len(events) < request_.pageSize:
                    break

    # Use stored checkpoint timestamp for start filter if applicable
    if checkpoint_name:
        checkpoint = cursor.get(checkpoint_name)
        if checkpoint:
            start = float(checkpoint)

    # Checks if start float timestamp was read from checkpoints
    start = parse_ts_to_posix_ts(start) if isinstance(start, str) else start

    request = QueryAuditLogRequest(
        actorIds=actor_ids.split(",") if actor_ids else None,
        actorIpAddresses=actor_ip_addresses.split(",") if actor_ip_addresses else None,
        actorNames=actor_names.split(",") if actor_names else None,
        eventTypes=event_types.split(",") if event_types else None,
        resourceIds=resource_ids.split(",") if resource_ids else None,
        userTypes=user_types.split(",") if user_types else None,
        dateRange=DateRange(
            startTime=start or None,
            endTime=parse_ts_to_posix_ts(end) if end else None,
        ),
        page=0,
        pageSize=100,
    )
    events_gen = yield_all_events(request)

    # Checkpointing
    if checkpoint_name:
        events_gen = _update_checkpoint(cursor, checkpoint_name, events_gen)

    with warn_interrupt() if checkpoint_name else nullcontext():
        if output:
            logger = get_server_logger(output, certs, ignore_cert_validation)
            for event in events_gen:
                logger.info(json.dumps(event))
            return

        if format_ == TableFormat.csv:
            render.csv(DefaultAuditEvent, events_gen, columns=columns, flat=True)
        else:
            printed = False
            for event in events_gen:  # Generator of audit log events in dict format
                printed = True
                if format_ == TableFormat.json_pretty:
                    console.print_json(data=event)
                else:
                    click.echo(json.dumps(event))
            if not printed:
                console.print("No results found.")


@audit_log.command()
@click.argument("checkpoint-name")
def clear_checkpoint(checkpoint_name: str):
    """Remove the saved audit log checkpoint from searches made with `--checkpoint` mode."""
    client = Client()
    cursor = _get_cursor_store(client.settings.api_client_id)
    cursor.delete(checkpoint_name)


@audit_log.command(cls=IncydrCommand)
@filter_options
@click.option(
    "--path",
    help="The file path where to save the CSV. Defaults to the current directory if not specified.",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    default=os.getcwd(),
)
@logging_options
def download(
    path: Optional[str],
    start: Optional[str],
    end: Optional[str],
    actor_ids: Optional[str],
    actor_ip_addresses: Optional[str],
    actor_names: Optional[str],
    event_types: Optional[str],
    resource_ids: Optional[str],
    user_types: Optional[str],
):
    """
    Download audit log events in CSV format.  Returns up to the most recent 100,000 events that match the search criteria.

    Use the --path option to specify where to save the CSV.  Defaults to the current directory if not specified.
    """
    client = Client()

    # convert str to list
    actor_ids = actor_ids.split(",") if isinstance(actor_ids, str) else None
    actor_names = actor_names.split(",") if isinstance(actor_names, str) else None
    actor_ip_addresses = (
        actor_ip_addresses.split(",") if isinstance(actor_ip_addresses, str) else None
    )
    event_types = event_types.split(",") if isinstance(event_types, str) else None
    resource_ids = resource_ids.split(",") if isinstance(resource_ids, str) else None
    user_types = user_types.split(",") if isinstance(user_types, str) else None

    client.audit_log.v1.download_events(
        path,
        actor_ids,
        actor_ip_addresses,
        actor_names,
        start,
        end,
        event_types,
        resource_ids,
        user_types,
    )

    console.print(f"Audit log events downloaded to '{path}'")


def _update_checkpoint(cursor, checkpoint_name, events_gen):
    """
    De-duplicates events across checkpointed runs.

    Since using the timestamp of the last event
    processed as the `--start` time of the next run causes the last event to show up again in the
    next results, we hash the last event(s) of each run and store those hashes in the cursor to
    filter out on the next run.

    It's also possible that two events have the exact same timestamp, so
    `checkpoint_events` needs to be a list of hashes so we can filter out everything that's actually
    been processed.
    """

    checkpoint_events = cursor.get_items(checkpoint_name)
    new_timestamp = None
    new_events = []
    for event in events_gen:
        # if event is a model, convert to a dict
        event_as_dict = event.dict() if isinstance(event, Model) else event
        event_hash = _hash_event(event_as_dict)
        if event_hash not in checkpoint_events:
            ts = parse_str_to_dt(event_as_dict["timestamp"])
            if not new_timestamp or ts > new_timestamp:
                new_timestamp = ts
                new_events.clear()
            new_events.append(event_hash)
            yield event
            new_timestamp.replace(tzinfo=timezone.utc)
            cursor.replace(checkpoint_name, new_timestamp.timestamp())
            cursor.replace_items(checkpoint_name, new_events)


def _hash_event(event):
    if isinstance(event, dict):
        event = json.dumps(event, sort_keys=True)
    return md5(event.encode()).hexdigest()


def _get_cursor_store(api_key):
    """
    Get cursor store for audit log search checkpoints.
    """
    dir_path = get_user_project_path(
        "checkpoints",
        api_key,
        "audit_log_checkpoints",
    )
    return CursorStore(dir_path, "audit_events")


class DefaultAuditEvent(Model):
    type: str = Field(None, alias="type$")
    actor_id: str = Field(None, alias="actorId")
    actor_name: str = Field(None, alias="actorName")
    actor_agent: str = Field(None, alias="actorAgent")
    actor_ip_addresses: str = Field(None, alias="actorIpAddress")
    timestamp: str
